import copy
import itertools

import geopandas.testing
import momepy
import numpy
import pandas
import pytest
import shapely
from pandas.testing import assert_series_equal

import neatnet

crs = "EPSG:3857"

point_1 = shapely.Point(1, 1)
point_2 = shapely.Point(2, 2)
point_3 = shapely.Point(3, 3)
point_4 = shapely.Point(4, 4)
point_5 = shapely.Point(5, 5)
point_6 = shapely.Point(6, 6)
point_7 = shapely.Point(7, 7)
point_8 = shapely.Point(8, 8)
point_9 = shapely.Point(9, 9)
point_24 = shapely.Point(2, 4)
point_42 = shapely.Point(4, 2)

split_list_2 = [point_2]
split_array_2 = numpy.array(split_list_2)
split_series_2 = geopandas.GeoSeries(split_array_2)

split_list_3 = [point_3]
split_array_3 = numpy.array(split_list_3)
split_series_3 = geopandas.GeoSeries(split_array_3)

split_list_2_3 = split_list_2 + split_list_3
split_array_2_3 = numpy.array(split_list_2_3)
split_series_2_3 = geopandas.GeoSeries(split_array_2_3)

split_list_2_8 = split_list_2 + [point_8]
split_array_2_8 = numpy.array(split_list_2_8)
split_series_2_8 = geopandas.GeoSeries(split_array_2_8)

split_list_2_3_7_8 = split_list_2_3 + [point_7, point_8]
split_array_2_3_7_8 = numpy.array(split_list_2_3_7_8)
split_series_2_3_7_8 = geopandas.GeoSeries(split_array_2_3_7_8)

line_1_4 = shapely.LineString((point_1, point_4))
line_1_2 = shapely.LineString((point_1, point_2))
line_1_3 = shapely.LineString((point_1, point_3))
line_2_3 = shapely.LineString((point_2, point_3))
line_2_4 = shapely.LineString((point_2, point_4))
line_3_4 = shapely.LineString((point_3, point_4))
line_6_9 = shapely.LineString((point_6, point_9))
line_6_7 = shapely.LineString((point_6, point_7))
line_6_8 = shapely.LineString((point_6, point_8))
line_7_8 = shapely.LineString((point_7, point_8))
line_8_9 = shapely.LineString((point_8, point_9))
line_24_42 = shapely.LineString((point_24, point_42))
line_24_3 = shapely.LineString((point_24, point_3))
line_3_42 = shapely.LineString((point_3, point_42))

cases = range(1, 9)
types = ["list", "array", "series"]

# case 1: 1 road input -- not split
cleaned_roads_1 = geopandas.GeoDataFrame(geometry=[line_1_2], crs=crs)
known_1 = cleaned_roads_1.copy()

# case 2: 1 road input -- split once
cleaned_roads_2 = geopandas.GeoDataFrame(geometry=[line_1_4], crs=crs)
known_2 = geopandas.GeoDataFrame(
    {"_status": ["changed", "changed"]},
    geometry=[line_1_2, line_2_4],
    crs=crs,
)

# case 3: 1 road input -- split twice
cleaned_roads_3 = geopandas.GeoDataFrame(geometry=[line_1_4], crs=crs)
known_3 = geopandas.GeoDataFrame(
    {"_status": ["changed", "changed", "changed"]},
    geometry=[line_1_2, line_2_3, line_3_4],
    crs=crs,
)

# case 4: 2 roads input -- neither roads split
cleaned_roads_4 = geopandas.GeoDataFrame(geometry=[line_1_2, line_2_4], crs=crs)
known_4 = cleaned_roads_4.copy()

# case 5: 2 roads input -- 1 road split once
cleaned_roads_5 = geopandas.GeoDataFrame(geometry=[line_1_4, line_6_9], crs=crs)
known_5 = geopandas.GeoDataFrame(
    {"_status": [numpy.nan, "changed", "changed"]},
    geometry=[line_6_9, line_1_2, line_2_4],
    crs=crs,
)

# case 6: 2 roads input -- 2 roads split once (unique splits)
cleaned_roads_6 = cleaned_roads_5.copy()
known_6 = geopandas.GeoDataFrame(
    {"_status": ["changed", "changed", "changed", "changed"]},
    geometry=[line_1_2, line_2_4, line_6_8, line_8_9],
    crs=crs,
)

# case 7: 2 roads input -- 2 roads split twice (unique splits)
cleaned_roads_7 = cleaned_roads_5.copy()
known_7 = geopandas.GeoDataFrame(
    {"_status": ["changed", "changed", "changed", "changed", "changed", "changed"]},
    geometry=[line_1_2, line_2_3, line_3_4, line_6_7, line_7_8, line_8_9],
    crs=crs,
)

# case 8: 2 roads input (perpendicular)-- 2 roads split once (intersection)
cleaned_roads_8 = geopandas.GeoDataFrame(geometry=[line_1_4, line_24_42], crs=crs)
known_8 = geopandas.GeoDataFrame(
    {"_status": ["changed", "changed", "changed", "changed"]},
    geometry=[line_1_3, line_3_4, line_24_3, line_3_42],
    crs=crs,
)


@pytest.mark.parametrize(
    "split_points,cleaned_roads,known",
    (
        [split_list_2, cleaned_roads_1, known_1],  # case 1
        [split_array_2, cleaned_roads_1, known_1],
        [split_series_2, cleaned_roads_1, known_1],
        [split_list_2, cleaned_roads_2, known_2],  # case 2
        [split_array_2, cleaned_roads_2, known_2],
        [split_series_2, cleaned_roads_2, known_2],
        [split_list_2_3, cleaned_roads_3, known_3],  # case 3
        [split_array_2_3, cleaned_roads_3, known_3],
        [split_series_2_3, cleaned_roads_3, known_3],
        [split_list_2, cleaned_roads_4, known_4],  # case 4
        [split_array_2, cleaned_roads_4, known_4],
        [split_series_2, cleaned_roads_4, known_4],
        [split_list_2, cleaned_roads_5, known_5],  # case 5
        [split_array_2, cleaned_roads_5, known_5],
        [split_series_2, cleaned_roads_5, known_5],
        [split_list_2_8, cleaned_roads_6, known_6],  # case 6
        [split_array_2_8, cleaned_roads_6, known_6],
        [split_series_2_8, cleaned_roads_6, known_6],
        [split_list_2_3_7_8, cleaned_roads_7, known_7],  # case 7
        [split_array_2_3_7_8, cleaned_roads_7, known_7],
        [split_series_2_3_7_8, cleaned_roads_7, known_7],
        [split_list_3, cleaned_roads_8, known_8],  # case 8
        [split_array_3, cleaned_roads_8, known_8],
        [split_series_3, cleaned_roads_8, known_8],
    ),
    ids=[f"case{c}-{t}" for c, t in list(itertools.product(cases, types))],
)
def test_split(split_points, cleaned_roads, known):
    observed = neatnet.nodes.split(split_points, cleaned_roads, crs)
    assert isinstance(observed, geopandas.GeoDataFrame)
    assert observed.crs == known.crs == cleaned_roads.crs == crs
    pytest.geom_test(observed.geometry, known.geometry)
    if "_status" in observed.columns:
        pandas.testing.assert_series_equal(observed["_status"], known["_status"])


point_20001 = shapely.Point(2.0001, 2.0001)
point_21 = shapely.Point(2.1, 2.1)

line_1_20001 = shapely.LineString((point_1, point_20001))
line_20001_4 = shapely.LineString((point_20001, point_4))
line_1_21 = shapely.LineString((point_1, point_21))
line_21_4 = shapely.LineString((point_21, point_4))
line_1_6 = shapely.LineString((point_1, point_6))


@pytest.mark.parametrize(
    "edge,split_point,tol,known",
    (
        [line_1_4, point_2, 0.0000001, numpy.array([line_1_2, line_2_4])],
        [line_1_4, point_20001, 0.0000001, numpy.array([line_1_20001, line_20001_4])],
        [line_1_4, point_21, 0.0001, numpy.array([line_1_21, line_21_4])],
        [line_1_4, point_6, 0.1, numpy.array([line_1_4])],
        [line_1_4, point_6, 3, numpy.array([line_1_6])],
    ),
    ids=["exact", "precise", "relaxed", "ignore", "extend"],
)
def test_snap_n_split(edge, split_point, tol, known):
    observed = neatnet.nodes._snap_n_split(edge, split_point, tol)
    numpy.testing.assert_array_equal(observed, known)


line_3_4 = shapely.LineString((point_3, point_4))
line_4_5 = shapely.LineString((point_4, point_5))
line_234 = shapely.LineString((point_2, point_3, point_4))

edgeline_types_get_components = [
    [line_1_2, line_2_4],
    numpy.array([line_1_2, line_3_4]),
    geopandas.GeoSeries([line_1_2, line_234]),
    [line_1_2, line_2_4] + [line_4_5],
]

ignore_types_get_components = [
    None,
    point_2,
    [point_2],
    numpy.array([point_3]),
    geopandas.GeoSeries([point_3]),
]

cases_types_get_components = [
    list(c)
    for c in itertools.product(
        edgeline_types_get_components, ignore_types_get_components
    )
]

known_get_components = [
    [0, 0],
    [2.0, 3.0],
    [2.0, 3.0],
    [0, 0],
    [0, 0],
    [2.0, 3.0],
    [2.0, 3.0],
    [2.0, 3.0],
    [2.0, 3.0],
    [2.0, 3.0],
    [0, 0],
    [2.0, 3.0],
    [2.0, 3.0],
    [0, 0],
    [0, 0],
    [0, 0, 0],
    [1.0, 0.0, 0.0],
    [1.0, 0.0, 0.0],
    [0, 0, 0],
    [0, 0, 0],
]

cases_get_components = [
    (*arg12, arg3)
    for arg12, arg3 in list(
        zip(cases_types_get_components, known_get_components, strict=True)
    )
]


t1_get_components = ["list", "ndarray", "GeoSeries", "list"]
t2_get_components = ["NoneType", "Point", "list", "ndarray", "GeoSeries"]
case_ids_get_components = [
    "-".join(c) for c in itertools.product(t1_get_components, t2_get_components)
]


@pytest.mark.parametrize(
    "edgelines,ignore,known",
    cases_get_components,
    ids=case_ids_get_components,
)
def test_get_components(edgelines, ignore, known):
    observed = neatnet.nodes.get_components(edgelines, ignore=ignore)
    numpy.testing.assert_array_equal(observed, known)


line_124 = shapely.LineString((point_1, point_2, point_4))
line_1234 = shapely.LineString((point_1, point_2, point_3, point_4))
line_245 = shapely.LineString((point_2, point_4, point_5))
line_1245 = shapely.LineString((point_1, point_2, point_4, point_5))

known_weld_edges = [
    [line_124],
    [line_1_2, line_2_4],
    [line_1_2, line_2_4],
    [line_124],
    [line_124],
    [line_1_2, line_3_4],
    [line_1_2, line_3_4],
    [line_1_2, line_3_4],
    [line_1_2, line_3_4],
    [line_1_2, line_3_4],
    [line_1234],
    [line_1_2, line_234],
    [line_1_2, line_234],
    [line_1234],
    [line_1234],
    [line_1245],
    [line_245, line_1_2],
    [line_245, line_1_2],
    [line_1245],
    [line_1245],
]

cases_types_weld_edges = copy.deepcopy(cases_types_get_components)


cases_weld_edges = [
    (*arg12, arg3)
    for arg12, arg3 in list(zip(cases_types_weld_edges, known_weld_edges, strict=True))
]

case_ids_weld_edges = copy.deepcopy(case_ids_get_components)


@pytest.mark.parametrize(
    "edgelines,ignore,known",
    cases_weld_edges,
    ids=case_ids_weld_edges,
)
def test_weld_edges(edgelines, ignore, known):
    observed = neatnet.nodes.weld_edges(edgelines, ignore=ignore)
    numpy.testing.assert_array_equal(observed, known)


class TestInduceNodes:
    def setup_method(self):
        self.p10 = shapely.Point(1, 0)
        self.p20 = shapely.Point(2, 0)
        self.p201 = shapely.Point(2, 0.1)
        self.p30 = shapely.Point(3, 0)
        self.p40 = shapely.Point(4, 0)
        self.p21 = shapely.Point(2, 1)
        self.p41 = shapely.Point(4, 1)
        self.p251 = shapely.Point(2.5, 1)
        self.p215 = shapely.Point(2, 1.5)
        self.p315 = shapely.Point(3, 1.5)

        self.line1020 = shapely.LineString((self.p10, self.p20))
        self.line1030 = shapely.LineString((self.p10, self.p30))
        self.line20121 = shapely.LineString((self.p201, self.p21))
        self.line2021 = shapely.LineString((self.p20, self.p21))
        self.line2030 = shapely.LineString((self.p20, self.p30))
        self.line2040 = shapely.LineString((self.p20, self.p40))
        self.line215351251 = shapely.LineString((self.p215, self.p315, self.p251))
        self.line251215 = shapely.LineString((self.p251, self.p215))
        self.line25141 = shapely.LineString((self.p251, self.p41))
        self.line2514130 = shapely.LineString((self.p251, self.p41, self.p30))
        self.line3040 = shapely.LineString((self.p30, self.p40))
        self.line3021251 = shapely.LineString((self.p30, self.p21, self.p251))
        self.line3041 = shapely.LineString((self.p30, self.p41))
        self.line4130 = shapely.LineString((self.p41, self.p30))
        self.line413021 = shapely.LineString((self.p41, self.p30, self.p21))
        self.line41302141 = shapely.LineString((self.p41, self.p30, self.p21, self.p41))
        self.line30214130 = shapely.LineString((self.p30, self.p21, self.p41, self.p30))
        self.line215315251215 = shapely.LineString(
            (self.p215, self.p315, self.p251, self.p215)
        )
        self.line251215315251 = shapely.LineString(
            (self.p251, self.p215, self.p315, self.p251)
        )

    def test_induced_simple(self):
        known = geopandas.GeoDataFrame(
            {
                "geometry": [self.line2021, self.line1020, self.line2030],
                "_status": [numpy.nan, "changed", "changed"],
            }
        )
        edges = geopandas.GeoDataFrame(geometry=[self.line1030, self.line2021])
        observed = neatnet.nodes.induce_nodes(edges)
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_not_induced_simple(self):
        known = geopandas.GeoDataFrame(geometry=[self.line1030, self.line20121])
        edges = geopandas.GeoDataFrame(geometry=[self.line1030, self.line20121])
        observed = neatnet.nodes.induce_nodes(edges)
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_induced_complex(self):
        known = geopandas.GeoDataFrame(
            {
                "geometry": [
                    self.line2030,
                    self.line3040,
                    self.line4130,
                    self.line3021251,
                    self.line25141,
                    self.line215351251,
                    self.line251215,
                ],
                "_status": ["changed"] * 7,
            }
        )
        edges = geopandas.GeoDataFrame(
            geometry=[self.line2040, self.line41302141, self.line215315251215]
        )
        observed = neatnet.nodes.induce_nodes(edges)
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_not_induced_complex(self):
        known = geopandas.GeoDataFrame(
            {
                "geometry": [
                    self.line2030,
                    self.line3040,
                    self.line3021251,
                    self.line2514130,
                    self.line251215315251,
                ],
            }
        )
        edges = geopandas.GeoDataFrame(
            geometry=[
                self.line2030,
                self.line3040,
                self.line3021251,
                self.line2514130,
                self.line251215315251,
            ]
        )
        observed = neatnet.nodes.induce_nodes(edges)
        geopandas.testing.assert_geodataframe_equal(observed, known)


class TestIdentifyDegreeMismatch:
    def setup_method(self):
        self.p20 = shapely.Point(2, 0)
        self.p30 = shapely.Point(3, 0)
        self.p40 = shapely.Point(4, 0)
        self.p21 = shapely.Point(2, 1)
        self.p41 = shapely.Point(4, 1)

        self.line2040 = shapely.LineString((self.p20, self.p40))
        self.line413021 = shapely.LineString((self.p41, self.p30, self.p21))
        self.line41302141 = shapely.LineString((self.p41, self.p30, self.p21, self.p41))

        self.sindex_kws = {"predicate": "dwithin", "distance": 1e-4}

    def test_no_mismatch(self):
        known = geopandas.GeoSeries([])
        edges = geopandas.GeoDataFrame(geometry=[self.line2040, self.line413021])
        observed = neatnet.nodes._identify_degree_mismatch(edges, self.sindex_kws)
        geopandas.testing.assert_geoseries_equal(observed, known)

    def test_mismatch(self):
        known = geopandas.GeoSeries([self.p41], index=[2])
        edges = geopandas.GeoDataFrame(geometry=[self.line2040, self.line41302141])
        observed = neatnet.nodes._identify_degree_mismatch(edges, self.sindex_kws)
        geopandas.testing.assert_geoseries_equal(observed, known)


class TestMakesLoopContact:
    def setup_method(self):
        self.p20 = shapely.Point(2, 0)
        self.p30 = shapely.Point(3, 0)
        self.p40 = shapely.Point(4, 0)
        self.p21 = shapely.Point(2, 1)
        self.p41 = shapely.Point(4, 1)
        self.p251 = shapely.Point(2.5, 1)
        self.p215 = shapely.Point(2, 1.5)
        self.p315 = shapely.Point(3, 1.5)

        self.line2040 = shapely.LineString((self.p20, self.p40))
        self.line2030 = shapely.LineString((self.p20, self.p30))
        self.line3040 = shapely.LineString((self.p30, self.p40))
        self.line3041 = shapely.LineString((self.p30, self.p41))
        self.line413021 = shapely.LineString((self.p41, self.p30, self.p21))
        self.line41302141 = shapely.LineString((self.p41, self.p30, self.p21, self.p41))
        self.line30214130 = shapely.LineString((self.p30, self.p21, self.p41, self.p30))
        self.line215315251215 = shapely.LineString(
            (self.p215, self.p315, self.p251, self.p215)
        )
        self.line251215315251 = shapely.LineString(
            (self.p251, self.p215, self.p315, self.p251)
        )

        self.sindex_kws = {"predicate": "dwithin", "distance": 1e-4}

    def test_off_1_bad_order(self):
        known_off_loops = geopandas.GeoSeries([self.p30], index=[1])
        known_on_loops = geopandas.GeoSeries([])
        edges = geopandas.GeoDataFrame(geometry=[self.line2040, self.line41302141])
        observed_off_loops, observed_on_loops = neatnet.nodes._makes_loop_contact(
            edges, self.sindex_kws
        )
        geopandas.testing.assert_geoseries_equal(observed_off_loops, known_off_loops)
        geopandas.testing.assert_geoseries_equal(observed_on_loops, known_on_loops)

    def test_off_2_bad_order(self):
        known_off_loops = geopandas.GeoSeries([self.p30, self.p30], index=[1, 1])
        known_on_loops = geopandas.GeoSeries([])
        edges = geopandas.GeoDataFrame(
            geometry=[self.line2030, self.line3040, self.line41302141]
        )
        observed_off_loops, observed_on_loops = neatnet.nodes._makes_loop_contact(
            edges, self.sindex_kws
        )
        geopandas.testing.assert_geoseries_equal(observed_off_loops, known_off_loops)
        geopandas.testing.assert_geoseries_equal(observed_on_loops, known_on_loops)

    def test_off_1_good_order(self):
        known_off_loops = geopandas.GeoSeries([self.p30, self.p30], index=[0, 3])
        known_on_loops = geopandas.GeoSeries([])
        edges = geopandas.GeoDataFrame(geometry=[self.line2040, self.line30214130])
        observed_off_loops, observed_on_loops = neatnet.nodes._makes_loop_contact(
            edges, self.sindex_kws
        )
        geopandas.testing.assert_geoseries_equal(observed_off_loops, known_off_loops)
        geopandas.testing.assert_geoseries_equal(observed_on_loops, known_on_loops)

    def test_off_2_good_order(self):
        known_off_loops = geopandas.GeoSeries([self.p30] * 4, index=[0, 0, 3, 3])
        known_on_loops = geopandas.GeoSeries([])
        edges = geopandas.GeoDataFrame(
            geometry=[self.line2030, self.line3040, self.line30214130]
        )
        observed_off_loops, observed_on_loops = neatnet.nodes._makes_loop_contact(
            edges, self.sindex_kws
        )
        geopandas.testing.assert_geoseries_equal(observed_off_loops, known_off_loops)
        geopandas.testing.assert_geoseries_equal(observed_on_loops, known_on_loops)

    def test_on_1_bad_order(self):
        known_off_loops = geopandas.GeoSeries([])
        known_on_loops = geopandas.GeoSeries([self.p251], index=[6])
        edges = geopandas.GeoDataFrame(
            geometry=[self.line41302141, self.line215315251215]
        )
        observed_off_loops, observed_on_loops = neatnet.nodes._makes_loop_contact(
            edges, self.sindex_kws
        )
        geopandas.testing.assert_geoseries_equal(observed_off_loops, known_off_loops)
        geopandas.testing.assert_geoseries_equal(observed_on_loops, known_on_loops)

    def test_on_1_good_order(self):
        known_off_loops = geopandas.GeoSeries([])
        known_on_loops = geopandas.GeoSeries([self.p251, self.p251], index=[4, 7])
        edges = geopandas.GeoDataFrame(
            geometry=[self.line41302141, self.line251215315251]
        )
        observed_off_loops, observed_on_loops = neatnet.nodes._makes_loop_contact(
            edges, self.sindex_kws
        )
        geopandas.testing.assert_geoseries_equal(observed_off_loops, known_off_loops)
        geopandas.testing.assert_geoseries_equal(observed_on_loops, known_on_loops)

    def test_multi_geoms(self):
        known_off_loops = geopandas.GeoSeries([self.p30], index=[1])
        known_on_loops = geopandas.GeoSeries([self.p251, self.p251], index=[4, 7])
        edges = geopandas.GeoDataFrame(
            geometry=[self.line2040, self.line41302141, self.line251215315251]
        )
        observed_off_loops, observed_on_loops = neatnet.nodes._makes_loop_contact(
            edges, self.sindex_kws
        )
        geopandas.testing.assert_geoseries_equal(observed_off_loops, known_off_loops)
        geopandas.testing.assert_geoseries_equal(observed_on_loops, known_on_loops)

    def test_no_loops(self):
        known_off_loops = geopandas.GeoSeries([])
        known_on_loops = geopandas.GeoSeries([])
        edges = geopandas.GeoDataFrame(geometry=[self.line2040, self.line3041])
        observed_off_loops, observed_on_loops = neatnet.nodes._makes_loop_contact(
            edges, self.sindex_kws
        )
        geopandas.testing.assert_geoseries_equal(observed_off_loops, known_off_loops)
        geopandas.testing.assert_geoseries_equal(observed_on_loops, known_on_loops)


class TestLoopsAndNonloops:
    def setup_method(self):
        self.p20 = shapely.Point(2, 0)
        self.p30 = shapely.Point(3, 0)
        self.p40 = shapely.Point(4, 0)
        self.p21 = shapely.Point(2, 1)
        self.p41 = shapely.Point(4, 1)
        self.p251 = shapely.Point(2.5, 1)
        self.p215 = shapely.Point(2, 1.5)
        self.p315 = shapely.Point(3, 1.5)

        self.line2040 = shapely.LineString((self.p20, self.p40))
        self.line413021 = shapely.LineString((self.p41, self.p30, self.p21))
        self.line41302141 = shapely.LineString((self.p41, self.p30, self.p21, self.p41))
        self.line251215315251 = shapely.LineString(
            (self.p215, self.p315, self.p251, self.p215)
        )

    def test_only_loops(self):
        known_loops = geopandas.GeoDataFrame(
            geometry=[self.line41302141, self.line251215315251]
        )
        known_non_loops = geopandas.GeoDataFrame(geometry=[])
        edges = geopandas.GeoDataFrame(
            geometry=[self.line41302141, self.line251215315251]
        )
        observed_loops, observed_non_loops = neatnet.nodes._loops_and_non_loops(edges)

        geopandas.testing.assert_geodataframe_equal(observed_loops, known_loops)
        geopandas.testing.assert_geodataframe_equal(observed_non_loops, known_non_loops)

    def test_both(self):
        known_loops = geopandas.GeoDataFrame(geometry=[self.line41302141], index=[1])
        known_non_loops = geopandas.GeoDataFrame(geometry=[self.line2040])
        edges = geopandas.GeoDataFrame(geometry=[self.line2040, self.line41302141])
        observed_loops, observed_non_loops = neatnet.nodes._loops_and_non_loops(edges)

        geopandas.testing.assert_geodataframe_equal(observed_loops, known_loops)
        geopandas.testing.assert_geodataframe_equal(observed_non_loops, known_non_loops)

    def test_only_non_loops(self):
        known_loops = geopandas.GeoDataFrame(geometry=[])
        known_non_loops = geopandas.GeoDataFrame(
            geometry=[self.line2040, self.line413021]
        )
        edges = geopandas.GeoDataFrame(geometry=[self.line2040, self.line413021])
        observed_loops, observed_non_loops = neatnet.nodes._loops_and_non_loops(edges)

        geopandas.testing.assert_geodataframe_equal(observed_loops, known_loops)
        geopandas.testing.assert_geodataframe_equal(observed_non_loops, known_non_loops)


class TestRemoveFalseNodes:
    def setup_method(self):
        p10 = shapely.Point(1, 0)
        p20 = shapely.Point(2, 0)
        p30 = shapely.Point(3, 0)
        p40 = shapely.Point(4, 0)
        p50 = shapely.Point(5, 0)
        p21 = shapely.Point(2, 1)
        p32 = shapely.Point(3, 2)
        p41 = shapely.Point(4, 1)

        self.line1020 = shapely.LineString((p10, p20))
        self.line2030 = shapely.LineString((p20, p30))
        self.line3040 = shapely.LineString((p30, p40))
        self.line4050 = shapely.LineString((p40, p50))
        self.line3021 = shapely.LineString((p30, p21))
        self.line2132 = shapely.LineString((p21, p32))
        self.line4132 = shapely.LineString((p41, p32))
        self.line3041 = shapely.LineString((p30, p41))
        self.attrs = ["cat"] * 3 + ["dog"] * 3 + ["eel"] * 2

        self.series = geopandas.GeoSeries(
            [
                self.line1020,
                self.line2030,
                self.line3040,
                self.line4050,
                self.line3021,
                self.line2132,
                self.line4132,
                self.line3041,
            ]
        )

        self.line102030 = shapely.LineString((p10, p20, p30))
        self.line304050 = shapely.LineString((p30, p40, p50))
        self.line3041323130 = shapely.LineString((p30, p41, p32, p21, p30))

        self.known_geoms = [
            self.line102030,
            self.line3041323130,
            self.line304050,
        ]

    def test_single_series(self):
        one_in_series = self.series[:0].copy()
        known = one_in_series
        observed = neatnet.nodes.remove_false_nodes(one_in_series)
        geopandas.testing.assert_geoseries_equal(observed, known)

    def test_series(self):
        known = geopandas.GeoDataFrame(geometry=self.known_geoms)
        observed = neatnet.nodes.remove_false_nodes(self.series)
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_frame(self):
        known = geopandas.GeoDataFrame(geometry=self.known_geoms)
        observed = neatnet.nodes.remove_false_nodes(
            geopandas.GeoDataFrame(geometry=self.series)
        )
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_frame_attrs_first(self):
        known = geopandas.GeoDataFrame(
            {"animal": ["cat", "dog", "cat"]},
            geometry=self.known_geoms,
            columns=["geometry", "animal"],
        )
        observed = neatnet.nodes.remove_false_nodes(
            geopandas.GeoDataFrame({"animal": self.attrs}, geometry=self.series)
        )
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_frame_attrs_last(self):
        known = geopandas.GeoDataFrame(
            {"animal": ["cat", "eel", "dog"]},
            geometry=self.known_geoms,
            columns=["geometry", "animal"],
        )
        observed = neatnet.nodes.remove_false_nodes(
            geopandas.GeoDataFrame({"animal": self.attrs}, geometry=self.series),
            aggfunc="last",
        )
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_momepy_suite(self):
        false_network = geopandas.read_file(
            momepy.datasets.get_path("tests"), layer="network"
        )
        false_network["vals"] = range(len(false_network))
        fixed = neatnet.remove_false_nodes(false_network).reset_index(drop=True)
        assert len(fixed) == 56
        assert isinstance(fixed, geopandas.GeoDataFrame)
        assert false_network.crs.equals(fixed.crs)
        assert sorted(false_network.columns) == sorted(fixed.columns)

        # check loop order
        expected = numpy.array(
            [
                [-727238.49292668, -1052817.28071986],
                [-727253.1752498, -1052827.47329062],
                [-727223.93217677, -1052829.47624082],
                [-727238.49292668, -1052817.28071986],
            ]
        )
        numpy.testing.assert_almost_equal(
            numpy.array(fixed.loc[55].geometry.coords), expected
        )

        fixed_series = neatnet.nodes.remove_false_nodes(
            false_network.geometry
        ).reset_index(drop=True)
        assert len(fixed_series) == 56
        assert isinstance(fixed_series, geopandas.GeoDataFrame)
        assert false_network.crs.equals(fixed_series.crs)

        multiindex = false_network.explode(index_parts=True)
        fixed_multiindex = neatnet.nodes.remove_false_nodes(multiindex)
        assert len(fixed_multiindex) == 56
        assert isinstance(fixed, geopandas.GeoDataFrame)
        assert sorted(false_network.columns) == sorted(fixed.columns)

        # no node of a degree 2
        df_streets = geopandas.read_file(
            momepy.datasets.get_path("bubenec"), layer="streets"
        )
        known = df_streets.drop([4, 7, 17, 22]).reset_index(drop=True)
        observed = neatnet.nodes.remove_false_nodes(known).reset_index(drop=True)
        geopandas.testing.assert_geodataframe_equal(observed, known)


class TestRotateLoopCoords:
    def setup_method(self):
        self.p20 = shapely.Point(2, 0)
        self.p30 = shapely.Point(3, 0)
        self.p40 = shapely.Point(4, 0)
        self.p21 = shapely.Point(2, 1)
        self.p41 = shapely.Point(4, 1)

        self.line2040 = shapely.LineString((self.p20, self.p40))
        self.line2030 = shapely.LineString((self.p20, self.p30))
        self.line3040 = shapely.LineString((self.p30, self.p40))
        self.line41302141 = shapely.LineString((self.p41, self.p30, self.p21, self.p41))
        self.line30214130 = shapely.LineString((self.p30, self.p21, self.p41, self.p30))

        self.known = numpy.array([[3.0, 0.0], [2.0, 1.0], [4.0, 1.0], [3.0, 0.0]])

    def test_needs_rotate_intersects_1(self):
        edges = geopandas.GeoDataFrame(geometry=[self.line2040, self.line41302141])
        observed = neatnet.nodes._rotate_loop_coords(
            edges[edges.is_ring].geometry,
            edges[~edges.is_ring],
        )
        numpy.testing.assert_array_equal(observed, self.known)

    def test_needs_rotate_intersects_2(self):
        edges = geopandas.GeoDataFrame(
            geometry=[self.line2030, self.line3040, self.line41302141]
        )
        observed = neatnet.nodes._rotate_loop_coords(
            edges[edges.is_ring].geometry,
            edges[~edges.is_ring],
        )
        numpy.testing.assert_array_equal(observed, self.known)

    def test_no_rotate_intersects_1(self):
        edges = geopandas.GeoDataFrame(geometry=[self.line2040, self.line30214130])
        observed = neatnet.nodes._rotate_loop_coords(
            edges[edges.is_ring].geometry,
            edges[~edges.is_ring],
        )
        numpy.testing.assert_array_equal(observed, self.known)

    def test_no_rotate_intersects_2(self):
        edges = geopandas.GeoDataFrame(
            geometry=[self.line2030, self.line3040, self.line30214130]
        )
        observed = neatnet.nodes._rotate_loop_coords(
            edges[edges.is_ring].geometry,
            edges[~edges.is_ring],
        )
        numpy.testing.assert_array_equal(observed, self.known)


def test_fix_topology():
    p20 = shapely.Point(2, 0)
    p30 = shapely.Point(3, 0)
    p40 = shapely.Point(4, 0)
    p21 = shapely.Point(2, 1)
    p41 = shapely.Point(4, 1)

    p251 = shapely.Point(2.5, 1)
    p215 = shapely.Point(2, 1.5)
    p315 = shapely.Point(3, 1.5)

    p27508 = shapely.Point(2.75, 0.8)
    p32508 = shapely.Point(3.25, 0.8)
    p31 = shapely.Point(3, 1)

    line2040 = shapely.LineString((p20, p40))
    line41313041 = shapely.LineString((p41, p21, p30, p41))
    line251215315251 = shapely.LineString((p215, p315, p251, p215))
    line275083250831 = shapely.LineString((p27508, p32508, p31, p27508))

    known = geopandas.GeoDataFrame(
        {
            "geometry": [
                shapely.LineString((p251, p215, p315, p251)),
                shapely.LineString((p31, p27508, p32508, p31)),
                shapely.LineString((p30, p41, p31)),
                shapely.LineString((p20, p30)),
                shapely.LineString((p30, p40)),
                shapely.LineString((p251, p21, p30)),
                shapely.LineString((p31, p251)),
            ],
            "_status": ["changed"] * 7,
        }
    )

    observed = neatnet.nodes.fix_topology(
        geopandas.GeoDataFrame(
            geometry=[
                line2040,
                line2040,
                line41313041,
                line251215315251,
                line275083250831,
            ]
        )
    )

    geopandas.testing.assert_geodataframe_equal(observed, known)


class TestConsolidateNodes:
    def setup_method(self):
        self.p1 = shapely.Point(10, 10)
        self.p2 = shapely.Point(15, 15)
        self.p3 = shapely.Point(15, 15.5)
        self.p4 = shapely.Point(17, 16)
        self.p5 = shapely.Point(17, 15)
        self.p6 = shapely.Point(10, 20)
        self.p7 = shapely.Point(20, 20)
        self.p8 = shapely.Point(20, 10)

        self.line1 = shapely.LineString((self.p1, self.p2))
        self.line2 = shapely.LineString((self.p2, self.p3))
        self.line3 = shapely.LineString((self.p3, self.p4))
        self.line4 = shapely.LineString((self.p4, self.p5))
        self.line5 = shapely.LineString((self.p5, self.p2))
        self.line6 = shapely.LineString((self.p6, self.p3))
        self.line7 = shapely.LineString((self.p7, self.p4))
        self.line8 = shapely.LineString((self.p8, self.p5))

        self.lines_array = numpy.array(
            [
                self.line1,
                self.line2,
                self.line3,
                self.line4,
                self.line5,
                self.line6,
                self.line7,
                self.line8,
            ]
        )

        self.lines_series = geopandas.GeoSeries(self.lines_array)
        self.lines_gframe = geopandas.GeoDataFrame(geometry=self.lines_series)

    def test_array_only_ends(self):
        known = geopandas.GeoDataFrame(
            {"geometry": numpy.array([self.line1]), "_status": ["original"]}
        )
        observed = neatnet.nodes.consolidate_nodes(geopandas.GeoSeries([self.line1]))
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_series_only_ends(self):
        known = geopandas.GeoDataFrame(
            {"geometry": [self.line1], "_status": ["original"]}
        )
        observed = neatnet.nodes.consolidate_nodes(geopandas.GeoSeries([self.line1]))
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_frame_only_ends(self):
        known = geopandas.GeoDataFrame(
            {"geometry": [self.line1], "_status": ["original"]}
        )
        observed = neatnet.nodes.consolidate_nodes(
            geopandas.GeoDataFrame(geometry=[self.line1]), preserve_ends=True
        )
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_array_no_change(self):
        known = geopandas.GeoDataFrame(
            {"geometry": self.lines_array, "_status": ["original"] * 8}
        )
        observed = neatnet.nodes.consolidate_nodes(self.lines_array, tolerance=0.1)
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_series_no_change(self):
        known = geopandas.GeoDataFrame(
            {"geometry": self.lines_series, "_status": ["original"] * 8}
        )
        observed = neatnet.nodes.consolidate_nodes(self.lines_series, tolerance=0.1)
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_frame_no_change(self):
        known = geopandas.GeoDataFrame(
            {"geometry": self.lines_series, "_status": ["original"] * 8}
        )
        observed = neatnet.nodes.consolidate_nodes(self.lines_gframe, tolerance=0.1)
        geopandas.testing.assert_geodataframe_equal(observed, known)

    def test_t05_pe(self):
        _p1 = shapely.Point(14.81439916202902, 15.667040754173883)
        _p2 = shapely.Point(15, 15.25)
        _p3 = shapely.Point(14.823223304703363, 14.823223304703363)
        _p4 = shapely.Point(15.242243505690334, 15.560560876422583)
        _p5 = shapely.Point(15.25, 15)

        known = geopandas.GeoDataFrame(
            {
                "geometry": [
                    shapely.LineString((self.p6, _p1, _p2)),
                    shapely.LineString((self.p1, _p3, _p2)),
                    shapely.LineString((_p2, _p4, self.p4)),
                    shapely.LineString((self.p5, _p5, _p2)),
                    shapely.LineString((self.p4, self.p5)),
                    shapely.LineString((self.p7, self.p4)),
                    shapely.LineString((self.p8, self.p5)),
                ],
                "_status": ["changed"] * 4 + ["original"] * 3,
            }
        )
        observed = neatnet.nodes.consolidate_nodes(
            self.lines_gframe, tolerance=0.5, preserve_ends=True
        )

        assert_series_equal(known._status, observed._status)
        pytest.geom_test(known, observed, tolerance=0.000001)

    def test_t1(self):
        _p1 = shapely.Point(14.628798324058037, 15.834081508347767)
        _p2 = shapely.Point(15, 15.25)
        _p3 = shapely.Point(14.646446609406727, 14.646446609406727)
        _p4 = shapely.Point(15.484487011380669, 15.621121752845166)
        _p5 = shapely.Point(16.5, 15.875)
        _p6 = shapely.Point(17, 15.5)
        _p7 = shapely.Point(16.5, 15)
        _p8 = shapely.Point(15.5, 15)
        _p9 = shapely.Point(17.256938038358054, 14.571769936069908)
        _p10 = shapely.Point(17.299642949358844, 16.39952393247846)

        known = geopandas.GeoDataFrame(
            {
                "geometry": [
                    shapely.LineString((self.p6, _p1, _p2)),
                    shapely.LineString((self.p1, _p3, _p2)),
                    shapely.LineString((_p2, _p4, _p5, _p6)),
                    shapely.LineString((_p6, _p7, _p8, _p2)),
                    shapely.LineString((self.p8, _p9, _p6)),
                    shapely.LineString((self.p7, _p10, _p6)),
                ],
                "_status": ["changed"] * 6,
            }
        )
        observed = neatnet.nodes.consolidate_nodes(self.lines_gframe, tolerance=1)

        assert_series_equal(known._status, observed._status)
        pytest.geom_test(known, observed, tolerance=0.000001)

    def test_t2(self):
        known = geopandas.GeoDataFrame(
            {
                "geometry": geopandas.GeoSeries.from_wkt(
                    [
                        "LINESTRING (10 20, 14.257597 16.168163, 15 15.25)",
                        "LINESTRING (10 10, 14.292893 14.292893, 15 15.25)",
                        "LINESTRING (15 15.25, 15.968974 15.742244, 16 15.75, 17 15.5)",
                        "LINESTRING (15 15.25, 16 15, 17 15.5)",
                        "LINESTRING (20 10, 17.513876 14.14354, 17 15.5)",
                        "LINESTRING (20 20, 17.599286 16.799048, 17 15.5)",
                    ]
                ),
                "_status": ["changed"] * 6,
            }
        )
        observed = neatnet.nodes.consolidate_nodes(self.lines_gframe, tolerance=2)

        assert_series_equal(known._status, observed._status)
        pytest.geom_test(known, observed, tolerance=0.000001)

    def test_t5(self):
        _p1 = shapely.Point(16, 15.375)
        _p2 = shapely.Point(13.143991620290183, 17.170407541738836)
        _p3 = shapely.Point(13.232233047033631, 13.232233047033631)
        _p4 = shapely.Point(18.28469019179028, 12.858849680349534)
        _p5 = shapely.Point(18.498214746794215, 17.99761966239229)

        known = geopandas.GeoDataFrame(
            {
                "geometry": [
                    shapely.LineString((self.p6, _p2, _p1)),
                    shapely.LineString((self.p1, _p3, _p1)),
                    shapely.LineString((self.p8, _p4, _p1)),
                    shapely.LineString((self.p7, _p5, _p1)),
                ],
                "_status": ["changed"] * 4,
            }
        )
        observed = neatnet.nodes.consolidate_nodes(self.lines_gframe, tolerance=5)

        assert_series_equal(known._status, observed._status)
        pytest.geom_test(known, observed, tolerance=0.000001)

    def test_t5_pe(self):
        _p1 = shapely.Point(16, 15.375)
        _p2 = shapely.Point(13.143991620290183, 17.170407541738836)
        _p3 = shapely.Point(13.232233047033631, 13.232233047033631)
        _p4 = shapely.Point(18.28469019179028, 12.858849680349534)
        _p5 = shapely.Point(18.498214746794215, 17.99761966239229)

        known = geopandas.GeoDataFrame(
            {
                "geometry": [
                    shapely.LineString((self.p6, _p2, _p1)),
                    shapely.LineString((self.p1, _p3, _p1)),
                    shapely.LineString((self.p8, _p4, _p1)),
                    shapely.LineString((self.p7, _p5, _p1)),
                ],
                "_status": ["changed"] * 4,
            }
        )
        observed = neatnet.nodes.consolidate_nodes(
            self.lines_gframe, tolerance=5, preserve_ends=True
        )

        assert_series_equal(known._status, observed._status)
        pytest.geom_test(known, observed, tolerance=0.000001)

    def test_t7(self):
        _p1 = shapely.Point(16.8, 16.3)
        _p2 = shapely.Point(12.40158826840626, 17.838570558434366)
        _p3 = shapely.Point(12.525126265847085, 12.525126265847085)
        _p4 = shapely.Point(18.79856626850639, 12.00238955248935)

        known = geopandas.GeoDataFrame(
            {
                "geometry": [
                    shapely.LineString((self.p6, _p2, _p1)),
                    shapely.LineString((self.p1, _p3, _p1)),
                    shapely.LineString((self.p8, _p4, _p1)),
                ],
                "_status": ["changed"] * 3,
            }
        )
        observed = neatnet.nodes.consolidate_nodes(self.lines_gframe, tolerance=7)

        assert_series_equal(known._status, observed._status)
        pytest.geom_test(known, observed, tolerance=0.000001)

    def test_t7_pe(self):
        _p1 = shapely.Point(16, 15.375)
        _p2 = shapely.Point(12.40158826840626, 17.838570558434366)
        _p3 = shapely.Point(12.525126265847085, 12.525126265847085)
        _p4 = shapely.Point(18.79856626850639, 12.00238955248935)
        _p5 = shapely.Point(19.097500645511904, 18.796667527349204)

        known = geopandas.GeoDataFrame(
            {
                "geometry": [
                    shapely.LineString((self.p6, _p2, _p1)),
                    shapely.LineString((self.p1, _p3, _p1)),
                    shapely.LineString((self.p8, _p4, _p1)),
                    shapely.LineString((self.p7, _p5, _p1)),
                ],
                "_status": ["changed"] * 4,
            }
        )
        observed = neatnet.nodes.consolidate_nodes(
            self.lines_gframe, tolerance=7, preserve_ends=True
        )

        assert_series_equal(known._status, observed._status)
        pytest.geom_test(known, observed, tolerance=0.000001)



"""Geometry-related functions"""

import collections
import math
import warnings

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely
from libpysal import graph
from scipy import spatial

from .nodes import consolidate_nodes


def _is_within(
    line: np.ndarray, poly: shapely.Polygon, rtol: float = 1e-4
) -> np.ndarray:
    """Check if the line is within a polygon with a set relative tolerance.

    Parameters
    ----------
    line : np.ndarray[shapely.LineString]
        Input line to check relationship.
    poly : shapely.Polygon
        Input polygon to check relationship.
    rtol : float = -1e4
        The set relative tolerance.

    Returns
    -------
    np.ndarray
        ``True`` if ``line`` is either entirely within ``poly`` or if
        ``line`` is within `poly`` based on a relaxed ``rtol`` relative tolerance.
    """

    within = shapely.within(line, poly)
    if within.all():
        return within

    intersection = shapely.intersection(line, poly)
    return np.abs(shapely.length(intersection) - shapely.length(line)) <= rtol


def angle_between_two_lines(
    line1: shapely.LineString, line2: shapely.LineString
) -> float:
    """Return the angle between 2 two lines (assuming they share a vertex).
    Based on ``momepy.coins`` but adapted to shapely lines.
    """

    return_bad = 0.0

    lines_distinct = line1 != line2
    if not lines_distinct:
        warnings.warn(
            f"Input lines are identical - must be distinct. Returning {return_bad}.",
            UserWarning,
            stacklevel=2,
        )
        return return_bad

    # extract points
    a, b, c, d = shapely.get_coordinates([line1, line2]).tolist()
    a, b, c, d = tuple(a), tuple(b), tuple(c), tuple(d)

    # assertion: we expect exactly 2 of the 4 points to be identical
    # (lines touch in this point)
    points = collections.Counter([a, b, c, d])

    lines_share_vertex = max(points.values()) > 1
    if not lines_share_vertex:
        warnings.warn(
            f"Input lines do not share a vertex. Returning {return_bad}.",
            UserWarning,
            stacklevel=2,
        )
        return return_bad

    # points where line touch = "origin" (for vector-based angle calculation)
    origin = [k for k, v in points.items() if v == 2][0]
    # other 2 unique points (one on each line)
    point1, point2 = (k for k, v in points.items() if v == 1)

    # translate lines into vectors (numpy arrays)
    v1 = [point1[0] - origin[0], point1[1] - origin[1]]
    v2 = [point2[0] - origin[0], point2[1] - origin[1]]

    # compute angle between 2 vectors in degrees
    dot_product = v1[0] * v2[0] + v1[1] * v2[1]
    norm_v1 = math.sqrt(v1[0] ** 2 + v1[1] ** 2)
    norm_v2 = math.sqrt(v2[0] ** 2 + v2[1] ** 2)
    cos_theta = round(dot_product / (norm_v1 * norm_v2), 6)  # precision issues fix
    angle = math.degrees(math.acos(cos_theta))

    return angle


def voronoi_skeleton(
    lines: list | np.ndarray | gpd.GeoSeries,
    poly: None | shapely.Polygon = None,
    snap_to: None | gpd.GeoSeries = None,
    max_segment_length: float | int = 1,
    buffer: None | float | int = None,
    secondary_snap_to: None | gpd.GeoSeries = None,
    clip_limit: None | float | int = 2,
    consolidation_tolerance: None | float | int = None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Returns average geometry.

    Parameters
    ----------
    lines : list | numpy.ndarray | geopandas.GeoSeries
        LineStrings connected at endpoints. If ``poly`` is passed in, ``lines``
        must be a ``geopandas.GeoSeries``.
    poly : None | shapely.Polygon = None
        Polygon enclosed by ``lines``.
    snap_to : None | gpd.GeoSeries = None
        Series of geometries that shall be connected to the skeleton.
    max_segment_length: float | int = 1
        Additional vertices will be added so that all line segments
        are no longer than this value. Must be greater than 0.
    buffer : None | float | int = None
        Optional custom buffer distance for dealing with Voronoi infinity issues.
    secondary_snap_to : None | gpd.GeoSeries = None
        Fall-back series of geometries that shall be connected to the skeleton.
    clip_limit : None | float | int = 2
        Following generation of the Voronoi linework, we clip to fit inside the polygon.
        To ensure we get a space to make proper topological connections from the
        linework to the actual points on the edge of the polygon, we clip using a
        polygon with a negative buffer of ``clip_limit`` or the radius of
        maximum inscribed circle, whichever is smaller.
    consolidation_tolerance : None | float | int = None
        Tolerance passed to node consolidation within the resulting skeleton.
        If ``None``, no consolidation happens.

    Returns
    -------
    edgelines : numpy.ndarray
        Array of averaged geometries.
    splitters : numpy.ndarray
        Split points.
    """
    if buffer is None:
        buffer = max_segment_length * 20
    if not poly:
        if not isinstance(lines, gpd.GeoSeries):
            lines = gpd.GeoSeries(lines)
        poly = shapely.box(*lines.total_bounds)
    # get an additional line around the lines to avoid infinity issues with Voronoi
    extended_lines = list(lines) + [poly.buffer(buffer).boundary]

    # interpolate lines to represent them as points for Voronoi
    shapely_lines = extended_lines
    points, ids = shapely.get_coordinates(
        shapely.segmentize(shapely_lines, max_segment_length), return_index=True
    )

    # remove duplicated coordinates
    unq, count = np.unique(points, axis=0, return_counts=True)
    mask = np.isin(points, unq[count > 1]).all(axis=1)
    points = points[~mask]
    ids = ids[~mask]

    # generate Voronoi diagram
    voronoi_diagram = spatial.Voronoi(points)

    # get all rigdes and filter only those between the two lines
    pts = voronoi_diagram.ridge_points
    mapped = np.take(ids, pts)
    rigde_vertices = np.array(voronoi_diagram.ridge_vertices)

    # iterate over segment-pairs and keep rigdes between input geometries
    _edgelines = []
    to_add = []
    splitters = []

    # determine the negative buffer distance to avoid overclipping of narrow polygons
    # this can still result in some missing links, but only in rare cases
    dist = min([clip_limit, shapely.ops.polylabel(poly).distance(poly.boundary) * 0.4])
    limit = poly.buffer(-dist)

    # drop ridges that are between points coming from the same line
    selfs = mapped[:, 0] == mapped[:, 1]
    buff = (mapped == mapped.max()).any(axis=1)
    mapped = mapped[~(selfs | buff)]
    rigde_vertices = rigde_vertices[~(selfs | buff)]
    unique = np.unique(np.sort(mapped, axis=1), axis=0)

    for a, b in unique:
        mask = ((mapped[:, 0] == a) | (mapped[:, 0] == b)) & (
            (mapped[:, 1] == a) | (mapped[:, 1] == b)
        )

        verts = rigde_vertices[mask]

        # generate the line in between the lines
        edgeline = shapely.line_merge(
            shapely.multilinestrings(voronoi_diagram.vertices[verts])
        )

        # check if the edgeline is within polygon
        if not edgeline.within(limit):
            # if not, clip it by the polygon with a small negative buffer to keep
            # the gap between edgeline and poly boundary to avoid possible
            # overlapping lines
            edgeline = shapely.intersection(edgeline, limit)

            # in edge cases, this can result in a MultiLineString with one sliver part
            edgeline = _remove_sliver(edgeline)

        # check if a, b lines share a node
        intersection = shapely_lines[b].intersection(shapely_lines[a])
        # if they do, add shortest line from the edgeline to the shared node and
        # combine it with the edgeline. Also, avoid an inner loop in more complex input
        # that would create connection across
        if not intersection.is_empty and not (
            intersection.geom_type == "MultiPoint"
            and (len(intersection.geoms) == 2 and len(lines) != 2)
        ):
            # we need union of edgeline and shortest because snap is buggy in GEOS
            # and line_merge as well. This results in a MultiLineString but we can
            # deal with those later. For now, we just need this extended edgeline to
            # be a single geometry to ensure the component discovery below works as
            # intended
            # get_parts is needed as in case of voronoi based on two lines, these
            # intersect on both ends, hence both need to be extended
            edgeline = shapely.union(
                edgeline,
                shapely.union_all(
                    shapely.shortest_line(
                        shapely.get_parts(intersection), edgeline.boundary
                    )
                ),
            )
        # add final edgeline to the list
        _edgelines.append(edgeline)

    edgelines = np.array(_edgelines)[~(shapely.is_empty(_edgelines))]

    if edgelines.shape[0] > 0:
        # if there is no explicit snapping target, snap to the boundary of the polygon
        # via the shortest line. That is by definition always within the polygon
        # (Martin thinks) (James concurs)
        if snap_to is not False:
            if snap_to is None:
                sl = shapely.shortest_line(
                    shapely.union_all(edgelines).boundary, poly.boundary
                )
                to_add.append(sl)
                splitters.append(shapely.get_point(sl, -1))

            # if we have some snapping targets, we need to figure out
            # what shall be snapped to what
            else:
                additions, splits = snap_to_targets(
                    edgelines, poly, snap_to, secondary_snap_to
                )
                to_add.extend(additions)
                splitters.extend(splits)

            # concatenate edgelines and their additions snapping to edge
            edgelines = np.concatenate([edgelines, to_add])
        # simplify to avoid unnecessary point density and some wobbliness
        edgelines = shapely.simplify(edgelines, max_segment_length)
    # drop empty
    edgelines = edgelines[edgelines != None]  # noqa: E711

    edgelines = shapely.line_merge(edgelines[shapely.length(edgelines) > 0])
    edgelines = _as_parts(edgelines)
    edgelines = _consolidate(edgelines, consolidation_tolerance)

    return edgelines, np.array(splitters)


def _remove_sliver(
    edgeline: shapely.LineString | shapely.MultiLineString,
) -> shapely.LineString:
    """Remove sliver(s) if present."""
    if edgeline.geom_type == "MultiLineString":
        parts = shapely.get_parts(edgeline)
        edgeline = parts[np.argmax(shapely.length(parts))]
    return edgeline


def _as_parts(edgelines: np.ndarray) -> np.ndarray:
    """Return constituent LineStrings if MultiLineString present."""
    if np.unique(shapely.get_type_id(edgelines)).shape[0] > 1:
        edgelines = shapely.get_parts(edgelines)
    return edgelines


def _consolidate(
    edgelines: np.ndarray, consolidation_tolerance: None | float | int
) -> np.ndarray:
    """Return ``edgelines`` from consolidated nodes, if criteria met."""
    if consolidation_tolerance and edgelines.shape[0] > 0:
        edgelines = consolidate_nodes(
            edgelines, tolerance=consolidation_tolerance, preserve_ends=True
        ).geometry.to_numpy()
    return edgelines


def snap_to_targets(
    edgelines: np.ndarray,
    poly: shapely.Polygon,
    snap_to: gpd.GeoSeries,
    secondary_snap_to: None | gpd.GeoSeries = None,
) -> tuple[list[shapely.LineString], list[shapely.Point]]:
    """Snap edgelines to vertices.

    Parameters
    ----------
    edgelines : numpy.ndarray
        Voronoi skeleton edges.
    poly : None | shapely.Polygon = None
        Polygon enclosed by ``lines``.
    snap_to : None | gpd.GeoSeries = None
        Series of geometries that shall be connected to the skeleton.
    secondary_snap_to : None | gpd.GeoSeries = None
        Fall-back series of geometries that shall be connected to the skeleton.

    Returns
    -------
    to_add, to_split : tuple[list[shapely.LineString], list[shapely.Point]]
        Lines to add and points where to split.
    """

    to_add: list = []
    to_split: list = []

    # generate graph from lines
    comp_labels, comp_counts, components = _prep_components(edgelines)

    primary_union = shapely.union_all(snap_to)
    secondary_union = shapely.union_all(secondary_snap_to)

    # if there are muliple components, loop over all and treat each
    if len(components) > 1:
        for comp_label, comp in components.geometry.items():
            cbound = comp.boundary

            # if component does not interest the boundary, it needs to be snapped
            # if it does but has only one part, this part interesect only on one
            # side (the node remaining from the removed edge) and needs to be
            # snapped on the other side as well
            if (
                (not comp.intersects(poly.boundary))
                or comp_counts[comp_label] == 1
                or (
                    not comp.intersects(primary_union)
                )  # ! this fixes one thing but may break others
            ):
                # add segment composed of the shortest line to the nearest snapping
                # target. We use boundary to snap to endpoints of edgelines only
                sl = shapely.shortest_line(cbound, primary_union)
                if _is_within(sl, poly):
                    to_split, to_add = _split_add(sl, to_split, to_add)
                else:
                    if secondary_snap_to is not None:
                        sl = shapely.shortest_line(cbound, secondary_union)
                        to_split, to_add = _split_add(sl, to_split, to_add)
    else:
        # if there is a single component, ensure it gets a shortest line to an
        # endpoint from each snapping target
        for target in snap_to:
            sl = shapely.shortest_line(components.boundary.item(), target)
            if _is_within(sl, poly):
                to_split, to_add = _split_add(sl, to_split, to_add)
            else:
                warnings.warn(
                    "Could not create a connection as it would lead outside "
                    "of the artifact.",
                    UserWarning,
                    stacklevel=2,
                )
    return to_add, to_split


def _prep_components(
    lines: np.ndarray | gpd.GeoSeries,
) -> tuple[pd.Series, pd.Series, gpd.GeoSeries]:
    """Helper for preparing graph components & labels in PySAL."""

    # cast edgelines to gdf
    lines = gpd.GeoDataFrame(geometry=lines)

    # build queen contiguity on edgelines and extract component labels
    not_empty = ~lines.is_empty
    not_nan = ~lines.geometry.isna()
    lines = lines[not_empty | not_nan]
    comp_labels = graph.Graph.build_contiguity(lines, rook=False).component_labels

    # compute size of each component
    comp_counts = comp_labels.value_counts()

    # get MultiLineString geometry per connected component
    components = lines.dissolve(comp_labels)

    return comp_labels, comp_counts, components


def _split_add(line: shapely.LineString, splits: list, adds: list) -> tuple[list, list]:
    """Helper for preparing splitter points & added lines."""
    splits.append(shapely.get_point(line, -1))
    adds.append(line)
    return splits, adds



import typing

import geopandas as gpd
import momepy
import networkx as nx
import numpy as np
import pandas as pd
import pyproj
import shapely
from scipy import sparse


def split(
    split_points: list | np.ndarray | gpd.GeoSeries,
    cleaned_roads: gpd.GeoDataFrame,
    crs: str | pyproj.CRS,
    eps: float = 1e-4,
) -> gpd.GeoSeries | gpd.GeoDataFrame:
    """Split lines on new nodes.

    Parameters
    ----------
    split_points : list | numpy.ndarray
        Points to split the ``cleaned_roads``.
    cleaned_roads : geopandas.GeoDataFrame
        Line geometries to be split with ``split_points``.
    crs : str | pyproj.CRS
        Anything accepted by ``pyproj.CRS``.
    eps : float = 1e-4
        Tolerance epsilon for point snapping.

    Returns
    -------
    geopandas.GeoSeries | geopandas.GeoDataFrame
        Resultant split line geometries.
    """
    split_points = gpd.GeoSeries(split_points, crs=crs)
    for split in split_points.drop_duplicates():
        _, ix = cleaned_roads.sindex.nearest(split, max_distance=eps)
        row = cleaned_roads.iloc[ix]
        edge = row.geometry
        if edge.shape[0] == 1:
            row = row.iloc[0]
            lines_split = _snap_n_split(edge.item(), split, eps)
            if lines_split.shape[0] > 1:
                gdf_split = gpd.GeoDataFrame(geometry=lines_split, crs=crs)
                for c in row.index.drop(["geometry", "_status"], errors="ignore"):
                    gdf_split[c] = row[c]
                gdf_split["_status"] = "changed"
                cleaned_roads = pd.concat(
                    [cleaned_roads.drop(edge.index[0]), gdf_split],
                    ignore_index=True,
                )
        elif edge.shape[0] > 1:
            to_be_dropped = []
            to_be_added = []
            for i, e in edge.items():
                lines_split = _snap_n_split(e, split, eps)
                if lines_split.shape[0] > 1:
                    to_be_dropped.append(i)
                    to_be_added.append(lines_split)

            if to_be_added:
                gdf_split = pd.DataFrame(
                    {"geometry": to_be_added, "_orig": to_be_dropped}
                ).explode("geometry")
                gdf_split = pd.concat(
                    [
                        gdf_split.drop(columns="_orig").reset_index(drop=True),
                        row.drop(columns="geometry")
                        .loc[gdf_split["_orig"]]
                        .reset_index(drop=True),
                    ],
                    axis=1,
                )
                gdf_split["_status"] = "changed"
                cleaned_roads = pd.concat(
                    [cleaned_roads.drop(to_be_dropped), gdf_split],
                    ignore_index=True,
                )
                cleaned_roads = gpd.GeoDataFrame(
                    cleaned_roads, geometry="geometry", crs=crs
                )

    return cleaned_roads.reset_index(drop=True)


def _snap_n_split(e: shapely.LineString, s: shapely.Point, tol: float) -> np.ndarray:
    """Snap point to edge and return lines to split."""
    snapped = shapely.snap(e, s, tolerance=tol)
    _lines_split = shapely.get_parts(shapely.ops.split(snapped, s))
    return _lines_split[~shapely.is_empty(_lines_split)]


def _status(x: pd.Series) -> str:
    """Determine the status of edge line(s)."""
    if len(x) == 1:
        return x.iloc[0]
    return "changed"


def get_components(
    edgelines: list | np.ndarray | gpd.GeoSeries,
    ignore: None | gpd.GeoSeries = None,
) -> np.ndarray:
    """Associate edges with connected component labels and return.

    Parameters
    ----------
    edgelines : list | np.ndarray | gpd.GeoSeries
        Collection of line objects.
    ignore : None | gpd.GeoSeries = None
        Nodes to ignore when labeling components.

    Returns
    -------
    np.ndarray
        Edge connected component labels.

    Notes
    -----
    See [https://github.com/uscuni/neatnet/issues/56] for detailed explanation of
    output.
    """
    edgelines = np.array(edgelines)
    start_points = shapely.get_point(edgelines, 0)
    end_points = shapely.get_point(edgelines, -1)
    points = shapely.points(
        np.unique(
            shapely.get_coordinates(np.concatenate([start_points, end_points])), axis=0
        )
    )
    if ignore is not None:
        mask = np.isin(points, ignore)
        points = points[~mask]
    # query LineString geometry to identify points intersecting 2 geometries
    inp, res = shapely.STRtree(shapely.boundary(edgelines)).query(
        points, predicate="intersects"
    )
    unique, counts = np.unique(inp, return_counts=True)
    mask = np.isin(inp, unique[counts == 2])
    merge_res = res[mask]
    merge_inp = inp[mask]
    closed = np.arange(len(edgelines))[shapely.is_closed(edgelines)]
    mask = np.isin(merge_res, closed) | np.isin(merge_inp, closed)
    merge_res = merge_res[~mask]
    merge_inp = merge_inp[~mask]
    g = nx.Graph(list(zip((merge_inp * -1) - 1, merge_res, strict=True)))
    components = {
        i: {v for v in k if v > -1} for i, k in enumerate(nx.connected_components(g))
    }
    component_labels = {value: key for key in components for value in components[key]}
    labels = pd.Series(component_labels, index=range(len(edgelines)))

    max_label = len(edgelines) - 1 if pd.isna(labels.max()) else int(labels.max())
    filling = pd.Series(range(max_label + 1, max_label + len(edgelines) + 1))
    labels = labels.fillna(filling)

    return labels.values


def weld_edges(
    edgelines: list | np.ndarray | gpd.GeoSeries,
    ignore: None | gpd.GeoSeries = None,
) -> list | np.ndarray | gpd.GeoSeries:
    """Combine lines sharing an endpoint (if only 2 lines share that point).
    Lightweight version of ``remove_false_nodes()``.

    Parameters
    ----------
    edgelines : list | np.ndarray | gpd.GeoSeries
        Collection of line objects.
    ignore : None | gpd.GeoSeries = None
        Nodes to ignore when welding components.

    Returns
    -------
    list | np.ndarray | gpd.GeoSeries
        Resultant welded ``edgelines`` if more than 1 passed in, otherwise
        the original ``edgelines`` object.
    """
    if len(edgelines) < 2:
        return edgelines
    labels = get_components(edgelines, ignore=ignore)
    return (
        gpd.GeoSeries(edgelines)
        .groupby(labels)
        .agg(lambda x: shapely.line_merge(shapely.GeometryCollection(x.values)))
    ).tolist()


def induce_nodes(roads: gpd.GeoDataFrame, eps: float = 1e-4) -> gpd.GeoDataFrame:
    """Adding potentially missing nodes on intersections of individual LineString
    endpoints with the remaining network. The idea behind is that if a line ends
    on an intersection with another, there should be a node on both of them.

    Parameters
    ----------
    roads : geopandas.GeoDataFrame
        Input LineString geometries.
    eps : float = 1e-4
        Tolerance epsilon for point snapping passed into ``nodes.split()``.

    Returns
    -------
    geopandas.GeoDataFrame
        Updated ``roads`` with (potentially) added nodes.
    """

    sindex_kws = {"predicate": "dwithin", "distance": 1e-4}

    # identify degree mismatch cases
    nodes_degree_mismatch = _identify_degree_mismatch(roads, sindex_kws)

    # ensure loop topology cases:
    #   - loop nodes intersecting non-loops
    #   - loop nodes intersecting other loops
    nodes_off_loops, nodes_on_loops = _makes_loop_contact(roads, sindex_kws)

    # all nodes to induce
    nodes_to_induce = pd.concat(
        [nodes_degree_mismatch, nodes_off_loops, nodes_on_loops]
    )

    return split(nodes_to_induce.geometry, roads, roads.crs, eps=eps)


def _identify_degree_mismatch(
    edges: gpd.GeoDataFrame, sindex_kws: dict
) -> gpd.GeoSeries:
    """Helper to identify difference of observed vs. expected node degree."""
    nodes = momepy.nx_to_gdf(momepy.node_degree(momepy.gdf_to_nx(edges)), lines=False)
    nix, eix = edges.sindex.query(nodes.geometry, **sindex_kws)
    coo_vals = ([True] * len(nix), (nix, eix))
    coo_shape = (len(nodes), len(edges))
    intersects = sparse.coo_array(coo_vals, shape=coo_shape, dtype=np.bool_)
    nodes["expected_degree"] = intersects.sum(axis=1)
    return nodes[nodes["degree"] != nodes["expected_degree"]].geometry


def _makes_loop_contact(
    edges: gpd.GeoDataFrame, sindex_kws: dict
) -> tuple[gpd.GeoSeries, gpd.GeoSeries]:
    """Helper to identify:
    1. loop nodes intersecting non-loops
    2. loop nodes intersecting other loops
    """

    loops, not_loops = _loops_and_non_loops(edges)
    loop_points = shapely.points(loops.get_coordinates().values)
    loop_gdf = gpd.GeoDataFrame(geometry=loop_points, crs=edges.crs)
    loop_point_geoms = loop_gdf.geometry

    # loop points intersecting non-loops
    nodes_from_non_loops_ix, _ = not_loops.sindex.query(loop_point_geoms, **sindex_kws)

    # loop points intersecting other loops
    nodes_from_loops_ix, _ = loops.sindex.query(loop_point_geoms, **sindex_kws)
    loop_x_loop, n_loop_x_loop = np.unique(nodes_from_loops_ix, return_counts=True)
    nodes_from_loops_ix = loop_x_loop[n_loop_x_loop > 1]

    # tease out both varieties
    nodes_non_loops = loop_gdf.loc[nodes_from_non_loops_ix]
    nodes_loops = loop_gdf.loc[nodes_from_loops_ix]

    return nodes_non_loops.geometry, nodes_loops.geometry


def _loops_and_non_loops(
    edges: gpd.GeoDataFrame,
) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
    """Bifurcate edge gdf into loops and non-loops."""
    loop_mask = edges.is_ring
    not_loops = edges[~loop_mask]
    loops = edges[loop_mask]
    return loops, not_loops


def remove_false_nodes(
    gdf: gpd.GeoSeries | gpd.GeoDataFrame, aggfunc: str | dict = "first", **kwargs
) -> gpd.GeoSeries | gpd.GeoDataFrame:
    """Reimplementation of ``momepy.remove_false_nodes()`` that preserves attributes.

    Parameters
    ----------
    gdf : geopandas.GeoSeries | geopandas.GeoDataFrame
        Input edgelines process. If any edges are ``MultiLineString`` they
        will be exploded into constituent ``LineString`` components.
    aggfunc : str | dict = 'first'
        Aggregate function for processing non-spatial component.
    **kwargs
        Keyword arguments for ``aggfunc``.

    Returns
    -------
    geopandas.GeoSeries | geopandas.GeoDataFrame
       The original input ``gdf`` if only 1 edgeline, otherwise the processed
       edgeline without interstitial nodes.

    Notes
    -----
    Any 3D geometries are (potentially) downcast in loops.
    """

    def merge_geometries(block: gpd.GeoSeries) -> shapely.LineString:
        """Helper in processing the spatial component."""
        return shapely.line_merge(shapely.GeometryCollection(block.values))

    if len(gdf) < 2:
        return gdf

    if isinstance(gdf, gpd.GeoSeries):
        gdf = gdf.to_frame("geometry")

    gdf = gdf.explode(ignore_index=True)

    labels = get_components(gdf.geometry)

    # Process non-spatial component
    data = gdf.drop(labels=gdf.geometry.name, axis=1)
    aggregated_data = data.groupby(by=labels).agg(aggfunc, **kwargs)
    aggregated_data.columns = aggregated_data.columns.to_flat_index()

    # Process spatial component
    g = gdf.groupby(group_keys=False, by=labels)[gdf.geometry.name].agg(
        merge_geometries
    )
    aggregated_geometry = gpd.GeoDataFrame(g, geometry=gdf.geometry.name, crs=gdf.crs)

    # Recombine
    aggregated = aggregated_geometry.join(aggregated_data)

    # Derive nodes
    nodes = momepy.nx_to_gdf(
        momepy.node_degree(momepy.gdf_to_nx(aggregated[[aggregated.geometry.name]])),
        lines=False,
    )

    # Bifurcate edges into loops and non-loops
    loops, not_loops = _loops_and_non_loops(aggregated)

    # Ensure:
    #   - all loops have exactly 1 endpoint; and
    #   - that endpoint shares a node with an intersecting line
    fixed_loops = []
    fixed_index = []
    node_ix, loop_ix = loops.sindex.query(nodes.geometry, predicate="intersects")
    for ix in np.unique(loop_ix):
        loop_geom = loops.geometry.iloc[ix]
        target_nodes = nodes.geometry.iloc[node_ix[loop_ix == ix]]
        if len(target_nodes) == 2:
            new_sequence = _rotate_loop_coords(loop_geom, not_loops)
            fixed_loops.append(shapely.LineString(new_sequence))
            fixed_index.append(ix)

    aggregated.loc[loops.index[fixed_index], aggregated.geometry.name] = fixed_loops
    return aggregated.reset_index(drop=True)


def _rotate_loop_coords(
    loop_geom: shapely.LineString, not_loops: gpd.GeoDataFrame
) -> np.ndarray:
    """Rotate loop node coordinates if needed to ensure topology."""

    loop_coords = shapely.get_coordinates(loop_geom)
    loop_points = gpd.GeoDataFrame(geometry=shapely.points(loop_coords))
    loop_points_ix, _ = not_loops.sindex.query(
        loop_points.geometry, predicate="dwithin", distance=1e-4
    )

    new_start = loop_points.loc[loop_points_ix].geometry.mode().get_coordinates().values
    _coords_match = (loop_coords == new_start).all(axis=1)
    new_start_idx = np.where(_coords_match)[0].squeeze()

    rolled_coords = np.roll(loop_coords[:-1], -new_start_idx, axis=0)
    new_sequence = np.append(rolled_coords, rolled_coords[[0]], axis=0)
    return new_sequence


def fix_topology(
    roads: gpd.GeoDataFrame,
    eps: float = 1e-4,
    **kwargs,
) -> gpd.GeoDataFrame:
    """Fix road network topology. This ensures correct topology of the network by:

        1.  Adding potentially missing nodes...
                on intersections of individual LineString endpoints
                with the remaining network. The idea behind is that
                if a line ends on an intersection with another, there
                should be a node on both of them.
        2. Removing nodes of degree 2...
                that have no meaning in the network used within our framework.
        3. Removing duplicated geometries (irrespective of orientation).

    Parameters
    ----------
    roads : geopandas.GeoDataFrame
        Input LineString geometries.
    eps : float = 1e-4
        Tolerance epsilon for point snapping passed into ``nodes.split()``.
    **kwargs : dict
        Key word arguments passed into ``remove_false_nodes()``.

    Returns
    -------
    gpd.GeoDataFrame
        The input roads that now have fixed topology and are ready
        to proceed through the simplification algorithm.
    """
    roads = roads[~roads.geometry.normalize().duplicated()].copy()
    roads_w_nodes = induce_nodes(roads, eps=eps)
    return remove_false_nodes(roads_w_nodes, **kwargs)


def consolidate_nodes(
    gdf: gpd.GeoDataFrame,
    tolerance: float = 2.0,
    preserve_ends: bool = False,
) -> gpd.GeoSeries:
    """Return geometry with consolidated nodes.

    Replace clusters of nodes with a single node (weighted centroid
    of a cluster) and snap linestring geometry to it. Cluster is
    defined using hierarchical clustering with average linkage
    on coordinates cut at a cophenetic distance equal to ``tolerance``.

    The use of hierachical clustering avoids the chaining effect of a sequence
    of intersections within ``tolerance`` from each other that would happen with
    DBSCAN and similar solutions.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        GeoDataFrame with LineStrings (usually representing street network).
    tolerance : float = 2.0
        The maximum distance between two nodes for one to be considered
        as in the neighborhood of the other. Nodes within tolerance are
        considered a part of a single cluster and will be consolidated.
    preserve_ends : bool = False
        If ``True``, nodes of a degree 1 will be excluded from the consolidation.

    Returns
    -------
    geopandas.GeoSeries
        Updated input ``gdf`` of LineStrings with consolidated nodes.
    """
    from scipy.cluster import hierarchy

    if isinstance(gdf, gpd.GeoSeries):
        gdf = gdf.to_frame("geometry")
    elif isinstance(gdf, np.ndarray):
        gdf = gpd.GeoDataFrame(geometry=gdf)

    nodes = momepy.nx_to_gdf(momepy.node_degree(momepy.gdf_to_nx(gdf)), lines=False)

    if preserve_ends:
        # keep at least one meter of original geometry around each end
        ends = nodes[nodes["degree"] == 1].buffer(1)
        nodes = nodes[nodes["degree"] > 1].copy()

        # if all we have are ends, return the original
        # - this is generally when called from within ``geometry._consolidate()``
        if nodes.shape[0] < 2:
            gdf["_status"] = "original"
            return gdf

    # get clusters of nodes which should be consolidated
    linkage = hierarchy.linkage(nodes.get_coordinates(), method="average")
    nodes["lab"] = hierarchy.fcluster(linkage, tolerance, criterion="distance")
    unique, counts = np.unique(nodes["lab"], return_counts=True)
    actual_clusters = unique[counts > 1]
    change = nodes[nodes["lab"].isin(actual_clusters)]

    # no change needed, return the original
    if change.empty:
        gdf["_status"] = "original"
        return gdf

    gdf = gdf.copy()
    # get geometry
    geom = gdf.geometry.copy()
    status = pd.Series("original", index=geom.index)

    # loop over clusters, cut out geometry within tolerance / 2 and replace it
    # with spider-like geometry to the weighted centroid of a cluster
    spiders = []
    midpoints = []

    clusters = change.dissolve(change["lab"])

    # TODO: not optimal but avoids some MultiLineStrings but not all
    cookies = clusters.buffer(tolerance / 2).convex_hull

    if preserve_ends:
        cookies = cookies.to_frame().overlay(ends.to_frame(), how="difference")

    for cluster, cookie in zip(clusters.geometry, cookies.geometry, strict=True):
        inds = geom.sindex.query(cookie, predicate="intersects")
        pts = shapely.get_coordinates(geom.iloc[inds].intersection(cookie.boundary))
        if pts.shape[0] > 0:
            # TODO: this may result in MultiLineString - we need to avoid that
            # TODO: It is temporarily fixed by that explode in return
            geom.iloc[inds] = geom.iloc[inds].difference(cookie)

            status.iloc[inds] = "changed"
            midpoint = np.mean(shapely.get_coordinates(cluster), axis=0)
            midpoints.append(midpoint)
            mids = np.array([midpoint] * len(pts))

            spider = shapely.linestrings(
                np.array([pts[:, 0], mids[:, 0]]).T,
                y=np.array([pts[:, 1], mids[:, 1]]).T,
            )
            spiders.append(spider)

    gdf = gdf.set_geometry(geom)
    gdf["_status"] = status

    if spiders:
        # combine geometries
        geoms = np.hstack(spiders)
        gdf = pd.concat([gdf, gpd.GeoDataFrame(geometry=geoms, crs=geom.crs)])

    agg: dict[str, str | typing.Callable] = {"_status": _status}
    for c in gdf.columns.drop(gdf.active_geometry_name):
        if c != "_status":
            agg[c] = "first"
    return remove_false_nodes(
        gdf[~gdf.geometry.is_empty].explode(),
        # NOTE: this aggfunc needs to be able to process all the columns
        aggfunc=agg,
    )



import logging
import pathlib
import time

import geopandas

import neatnet

start_time = time.time()

logging.basicConfig(
    filename="simplified_generation.log",
    filemode="a",
    format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
    datefmt="%H:%M:%S",
    level=logging.NOTSET,
)
logger = logging.getLogger(__name__)

logging.info("")
logging.info("")
logging.info(" |‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾|")
logging.info(" | Generating Simplified Street Networks |")
logging.info(" |_______________________________________|")
logging.info("")
logging.info("")
logging.info("")

fua_city = {
    1133: "aleppo",
    869: "auckland",
    4617: "bucaramanga",
    809: "douala",
    1656: "liege",
    4881: "slc",
}

# dict of cityname: fua ID
city_fua = {c: f for f, c in fua_city.items()}

for city, fua in city_fua.items():
    t1 = time.time()
    aoi = f"{city}_{fua}"

    logging.info("")
    logging.info("")
    logging.info("")
    logging.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ >>>>")
    logging.info("")
    logging.info("")
    logging.info(f"  ** {aoi} **")
    logging.info("")
    logging.info("")

    # input data
    original = geopandas.read_parquet(pathlib.Path(aoi, "original.parquet"))

    # output data
    simplified = neatnet.simplify_network(original)
    simplified.to_parquet(pathlib.Path(aoi, "simplified.parquet"))

    t2 = round((time.time() - t1) / 60.0, 2)

    logging.info("")
    logging.info("")
    logging.info(f"\t{aoi} runtime: {t2} minutes")
    logging.info("")
    logging.info("")
    logging.info("")
    logging.info("<<<< ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
    logging.info("")

endtime_time = round((time.time() - start_time) / 60.0, 2)

logging.info("")
logging.info("")
logging.info(f"Total runtime: {endtime_time} minutes")
logging.info(
    "========================================================================="
)
logging.info("")
logging.info("")



import contextlib
from importlib.metadata import PackageNotFoundError, version

from . import simplify
from .artifacts import get_artifacts
from .nodes import (
    consolidate_nodes,
    fix_topology,
    induce_nodes,
    remove_false_nodes,
    split,
)
from .simplify import simplify_loop, simplify_network

with contextlib.suppress(PackageNotFoundError):
    __version__ = version("neatnet")



import geopandas.testing
import numpy
import pandas
import pytest
import shapely

import neatnet


class TestIsWithin:
    def setup_method(self):
        self.polygon = shapely.Polygon(((0, 0), (10, 0), (10, 10), (0, 10), (0, 0)))

    def test_within_fully(self):
        line = shapely.LineString(((2, 2), (8, 8)))

        known = True
        observed = neatnet.geometry._is_within(line, self.polygon)

        assert known == observed

    def test_within_tol(self):
        line = shapely.LineString(((2, 2), (2, 10.0001)))

        known = True
        observed = neatnet.geometry._is_within(line, self.polygon)

        assert known == observed

    def test_not_within_tol(self):
        line = shapely.LineString(((2, 2), (2, 10.001)))

        known = False
        observed = neatnet.geometry._is_within(line, self.polygon)

        assert known == observed

    def test_within_tol_strict(self):
        line = shapely.LineString(((2, 2), (2, 10.0000001)))

        known = True
        observed = neatnet.geometry._is_within(line, self.polygon, rtol=1e-7)

        assert known == observed

    def test_not_within_tol_strict(self):
        line = shapely.LineString(((2, 2), (2, 10.000001)))

        known = False
        observed = neatnet.geometry._is_within(line, self.polygon, rtol=1e-7)

        assert known == observed

    def test_within_tol_relaxed(self):
        line = shapely.LineString(((2, 2), (2, 11)))

        known = True
        observed = neatnet.geometry._is_within(line, self.polygon, rtol=1)

        assert known == observed

    def test_not_within_tol_relaxed(self):
        line = shapely.LineString(((2, 2), (2, 12)))

        known = False
        observed = neatnet.geometry._is_within(line, self.polygon, rtol=1)

        assert known == observed

    def test_not_within(self):
        line = shapely.LineString(((11, 11), (12, 12)))

        known = False
        observed = neatnet.geometry._is_within(line, self.polygon)

        assert known == observed


class TestAngleBetween2Lines:
    def setup_method(self):
        self.line1 = shapely.LineString(((0, 0), (1, 0)))
        self.line2 = shapely.LineString(((1, 0), (1, 1)))
        self.line3 = shapely.LineString(((0, 0), (0, 1)))
        self.line4 = shapely.LineString(((0, 1), (1, 1)))

    def test_q1(self):
        known = 90.0
        observed = neatnet.geometry.angle_between_two_lines(self.line1, self.line3)
        assert observed == known

    def test_q2(self):
        known = 90.0
        observed = neatnet.geometry.angle_between_two_lines(self.line1, self.line2)
        assert observed == known

    def test_q3(self):
        known = 90.0
        observed = neatnet.geometry.angle_between_two_lines(self.line2, self.line4)
        assert observed == known

    def test_q4(self):
        known = 90.0
        observed = neatnet.geometry.angle_between_two_lines(self.line3, self.line4)
        assert observed == known

    def test_indistinct(self):
        known = 0.0
        with pytest.warns(
            UserWarning,
            match="Input lines are identical - must be distinct. Returning 0.0.",
        ):
            observed = neatnet.geometry.angle_between_two_lines(self.line1, self.line1)
        assert observed == known

    def test_not_adjacent(self):
        known = 0.0
        with pytest.warns(
            UserWarning, match="Input lines do not share a vertex. Returning 0.0."
        ):
            observed = neatnet.geometry.angle_between_two_lines(self.line1, self.line4)
        assert observed == known


voronoi_skeleton_params = pytest.mark.parametrize(
    "lines_type,as_poly,buffer",
    [
        (list, False, None),
        (list, True, 0.001),
        (numpy.array, False, 0.01),
        (numpy.array, True, 0.1),
        (pandas.Series, False, 1),
        (pandas.Series, True, 2.0),
        (geopandas.GeoSeries, False, 5),
        (geopandas.GeoSeries, True, 10.314),
    ],
)


class TestVoronoiSkeleton:
    def setup_method(self):
        self.square = [
            shapely.LineString(((0, 0), (1000, 0))),
            shapely.LineString(((1000, 0), (1000, 1000))),
            shapely.LineString(((0, 0), (0, 1000))),
            shapely.LineString(((0, 1000), (1000, 1000))),
        ]
        self.known_square_skeleton_edges = numpy.array(
            [
                shapely.LineString(((1000, 0), (998, 2), (500, 500))),
                shapely.LineString(((0, 0), (2, 2), (500, 500))),
                shapely.LineString(((1000, 1000), (998, 998), (500, 500))),
                shapely.LineString(((0, 1000), (2, 998), (500, 500))),
            ]
        )
        self.known_square_skeleton_splits = [shapely.Point(0, 0)]
        self.known_square_skeleton_splits_snap_to = [
            shapely.Point(1000, 0),
            shapely.Point(0, 0),
            shapely.Point(0, 1000),
            shapely.Point(1000, 1000),
        ]

    @voronoi_skeleton_params
    def test_square(self, lines_type, as_poly, buffer):
        known_edges = self.known_square_skeleton_edges
        known_splits = self.known_square_skeleton_splits

        lines = lines_type(self.square)
        poly = pytest.polygonize(lines) if as_poly else None
        observed_edges, observed_splits = neatnet.geometry.voronoi_skeleton(
            lines,
            poly=poly,
            buffer=buffer,
        )

        pytest.geom_test(observed_edges, known_edges)
        pytest.geom_test(observed_splits, known_splits)

    @voronoi_skeleton_params
    def test_square_snap_to(self, lines_type, as_poly, buffer):
        known_edges = self.known_square_skeleton_edges
        known_splits = self.known_square_skeleton_splits_snap_to

        lines = lines_type(self.square)
        poly = pytest.polygonize(lines) if as_poly else None
        observed_edges, observed_splits = neatnet.geometry.voronoi_skeleton(
            lines,
            poly=poly,
            buffer=buffer,
            snap_to=(
                pytest.polygonize(geopandas.GeoSeries(lines), as_geom=False)
                .extract_unique_points()
                .explode()
            ),
        )

        pytest.geom_test(observed_edges, known_edges)
        pytest.geom_test(observed_splits, known_splits)


line_100_900 = shapely.LineString(((1000, 1000), (1000, 9000)))
line_100_120 = shapely.LineString(((1000, 1020), (1020, 1020)))
lines_100_900_100_120 = shapely.MultiLineString((line_100_900, line_100_120))
line_110_900 = shapely.LineString(((1000, 9000), (1100, 9000)))


def test_remove_sliver():
    known = line_100_900
    observed = neatnet.geometry._remove_sliver(lines_100_900_100_120)
    assert observed == known


def test_as_parts():
    known = numpy.array([line_100_900, line_100_120, line_110_900])
    observed = neatnet.geometry._as_parts(
        numpy.array([lines_100_900_100_120, line_110_900])
    )
    numpy.testing.assert_array_equal(observed, known)


@pytest.mark.parametrize("tolerance", [0.1, 1, 10, 100, 1_000, 10_000, 100_000])
def test_consolidate(tolerance):
    known = numpy.array([line_100_900, line_100_120, line_110_900])
    observed = neatnet.geometry._consolidate(
        numpy.array([line_100_900, line_100_120, line_110_900]), tolerance
    )
    numpy.testing.assert_array_equal(observed, known)


def test_prep_components():
    line1 = shapely.LineString(((1, 1), (1, 2)))
    line2 = shapely.LineString(((1, 2), (2, 2)))
    line3 = shapely.LineString(((3, 0), (3, 3)))

    known_labels = pandas.Series(
        [0, 0, 1],
        index=pandas.Index([0, 1, 2], name="focal"),
        name="component labels",
        dtype=numpy.int32,
    )
    known_counts = pandas.Series(
        [2, 1],
        index=pandas.Index([0, 1], name="component labels", dtype=numpy.int32),
        name="count",
        dtype=numpy.int64,
    )
    known_comps = geopandas.GeoDataFrame(
        geometry=[
            shapely.MultiLineString(
                (
                    shapely.LineString(((1, 1), (1, 2))),
                    shapely.LineString(((1, 2), (2, 2))),
                )
            ),
            shapely.LineString(((3, 0), (3, 3))),
        ],
        index=pandas.Index([0, 1], name="component labels", dtype=numpy.int32),
    )

    observed_labels, observed_counts, observed_comps = (
        neatnet.geometry._prep_components([line1, line2, line3])
    )

    pandas.testing.assert_series_equal(observed_labels, known_labels)
    pandas.testing.assert_series_equal(observed_counts, known_counts)
    geopandas.testing.assert_geodataframe_equal(observed_comps, known_comps)


def test_split_add():
    _x = 1100
    x1, y1 = _x, 0
    x2, y2 = _x, 1000
    sl = shapely.LineString(((x1, y1), (x2, y2)))
    known_splits = [shapely.Point((x2, y2))]
    known_adds = [sl]
    observed_splits, observed_adds = neatnet.geometry._split_add(sl, [], [])
    assert observed_splits == known_splits
    assert observed_adds == known_adds


class TestSnapToTargets:
    def setup_method(self):
        # edgelines
        line1 = shapely.LineString(((100, 100), (1000, 100)))
        line2 = shapely.LineString(((1000, 100), (1000, 1000)))
        line3 = shapely.LineString(((100, 100), (100, 1000)))
        line4 = shapely.LineString(((100, 1000), (1000, 1000)))
        self.lines = [line1, line2, line3, line4]

        # poly
        self.poly = shapely.polygonize(self.lines).buffer(0)

        # snap_to
        self.snap_to_1 = (
            geopandas.GeoSeries(self.lines)
            .polygonize()
            .extract_unique_points()
            .explode()
        )

    def test_warn(self):
        with pytest.warns(
            UserWarning,
            match=(
                "Could not create a connection as it would "
                "lead outside of the artifact."
            ),
        ):
            neatnet.geometry.snap_to_targets(
                self.lines,
                self.poly,
                snap_to=self.snap_to_1,
            )

    def test_secondary(self):
        known = ([None], [None])

        line1_b = shapely.LineString(((500, 500), (1500, 500)))
        line2_b = shapely.LineString(((1500, 500), (1500, 1500)))
        line3_b = shapely.LineString(((500, 500), (500, 1500)))
        line4_b = shapely.LineString(((500, 1500), (1500, 1500)))
        lines_b = [line1_b, line2_b, line3_b, line4_b]
        snap_to_2 = (
            geopandas.GeoSeries(lines_b).polygonize().extract_unique_points().explode()
        )

        observed = neatnet.geometry.snap_to_targets(
            self.lines + lines_b,
            self.poly,
            snap_to=self.snap_to_1,
            secondary_snap_to=snap_to_2,
        )

        assert observed == known



import platform

import geopandas.testing
import numpy
import pandas
import pytest
import shapely

import neatnet

# set the global exception raiser for testing & debugging
# See gh#121
neatnet.simplify.DEBUGGING = True

line_collection = (  # type: ignore[valid-type, misc]
    list[shapely.LineString]
    | tuple[shapely.LineString]
    | numpy.ndarray
    | pandas.Series
    | geopandas.GeoSeries
)

geometry_collection = (
    list[shapely.GeometryCollection]
    | tuple[shapely.GeometryCollection]
    | numpy.ndarray
    | pandas.Series
    | geopandas.GeoSeries
)


####################################################
# see:
#   - gh#106
#   - gh#102
#   - gh#77
#   - gh#75
#   - gh#74
KNOWN_BAD_GEOMS = {
    "aleppo_1133": [],
    "auckland_869": [1412],
    "bucaramanga_4617": [],
    "douala_809": [],
    "liege_1656": [921],
    "slc_4881": [1144, 1146],
    "apalachicola_standard": [324],
    "apalachicola_exclusion_mask": [],
}
####################################################


def polygonize(
    collection: line_collection,  # type: ignore[valid-type]
    as_geom: bool = True,  # type: ignore[valid-type]
) -> shapely.Polygon | geopandas.GeoSeries:
    """Testing helper -- Create polygon from collection of lines."""
    if isinstance(collection, pandas.Series | geopandas.GeoSeries):
        _poly = geopandas.GeoSeries(collection).polygonize()
        if as_geom:
            return _poly.squeeze()
        else:
            return _poly
    else:
        return shapely.polygonize(collection).buffer(0)


def is_geopandas(collection: geometry_collection) -> bool:  # type: ignore[valid-type]
    return isinstance(collection, geopandas.GeoSeries | geopandas.GeoDataFrame)


def geom_test(
    collection1: geometry_collection,  # type: ignore[valid-type]
    collection2: geometry_collection,  # type: ignore[valid-type]
    tolerance: float = 1e-1,
    aoi: None | str = None,
) -> bool:
    """Testing helper -- geometry verification."""

    if not is_geopandas(collection1):
        collection1 = geopandas.GeoSeries(collection1)

    if not is_geopandas(collection2):
        collection2 = geopandas.GeoSeries(collection2)

    geoms1 = collection1.geometry.normalize()  # type: ignore[attr-defined]
    geoms2 = collection2.geometry.normalize()  # type: ignore[attr-defined]

    if aoi and aoi.startswith("apalachicola"):
        # Varied index order across OSs.
        # See [https://github.com/uscuni/neatnet/pull/104#issuecomment-2495572388]
        geoms1 = geoms1.sort_values().reset_index(drop=True)
        geoms2 = geoms2.sort_values().reset_index(drop=True)

    try:
        assert shapely.equals_exact(geoms1, geoms2, tolerance=tolerance).all()
    except AssertionError:
        unexpected_bad = {}
        for ix in geoms1.index:
            g1 = geoms1.loc[ix]
            g2 = geoms2.loc[ix]
            if (
                not shapely.equals_exact(g1, g2, tolerance=tolerance)
                and ix not in KNOWN_BAD_GEOMS[aoi]  # type: ignore[index]
            ):
                unexpected_bad[ix] = {
                    "n_coords": {
                        "g1": shapely.get_coordinates(g1).shape[0],
                        "g2": shapely.get_coordinates(g2).shape[0],
                    },
                    "length": {"g1": g1.length, "g2": g2.length},
                }
        if unexpected_bad:
            raise AssertionError(
                f"Problem in '{aoi}' – check locs: {unexpected_bad}"
            ) from None
    return True


def pytest_addoption(parser):
    """Add custom command line arguments."""

    # flag for determining CI environment
    parser.addoption(
        "--env_type",
        action="store",
        default="latest",
        help="Testing environment type label",
        type=str,
    )


def pytest_configure(config):  # noqa: ARG001
    """PyTest session attributes, methods, etc."""

    valid_env_types = ["oldest", "latest", "dev"]
    pytest.env_type = config.getoption("env_type").split("_")[-1]
    assert pytest.env_type in valid_env_types

    pytest.ubuntu = "ubuntu" in platform.version().lower()

    pytest.polygonize = polygonize
    pytest.geom_test = geom_test



import logging
import typing
import warnings

import geopandas as gpd
import momepy
import numpy as np
import pandas as pd
import shapely
from libpysal import graph
from scipy import sparse

from .artifacts import (
    get_artifacts,
    n1_g1_identical,
    nx_gx,
    nx_gx_cluster,
    nx_gx_identical,
)
from .continuity import continuity, get_stroke_info
from .nodes import (
    _status,
    consolidate_nodes,
    fix_topology,
    induce_nodes,
    remove_false_nodes,
    split,
)

DEBUGGING = False

logger = logging.getLogger(__name__)


def _link_nodes_artifacts(
    step: str,
    roads: gpd.GeoDataFrame,
    artifacts: gpd.GeoDataFrame,
    eps: None | float,
) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
    """Helper to prep nodes & artifacts when simplifying singletons & pairs."""

    # Get nodes from the network
    nodes = momepy.nx_to_gdf(momepy.node_degree(momepy.gdf_to_nx(roads)), lines=False)

    if step == "singletons":
        node_geom = nodes.geometry
        sindex_kwargs = {"predicate": "dwithin", "distance": eps}
    else:
        node_geom = nodes.buffer(0.1)
        sindex_kwargs = {"predicate": "intersects"}

    # Link nodes to artifacts
    node_idx, artifact_idx = artifacts.sindex.query(node_geom, **sindex_kwargs)

    intersects = sparse.coo_array(
        ([True] * len(node_idx), (node_idx, artifact_idx)),
        shape=(len(nodes), len(artifacts)),
        dtype=np.bool_,
    )

    # Compute number of nodes per artifact
    artifacts["node_count"] = intersects.sum(axis=0)

    return nodes, artifacts


def _classify_strokes(
    artifacts: gpd.GeoDataFrame, roads: gpd.GeoDataFrame
) -> gpd.GeoDataFrame:
    """Classify artifacts with ``{C,E,S}`` typology."""

    strokes, c_, e_, s_ = get_stroke_info(artifacts, roads)

    artifacts["stroke_count"] = strokes
    artifacts["C"] = c_
    artifacts["E"] = e_
    artifacts["S"] = s_

    return artifacts


def _identify_non_planar(
    artifacts: gpd.GeoDataFrame, roads: gpd.GeoDataFrame
) -> gpd.GeoDataFrame:
    """Filter artifacts caused by non-planar intersections."""

    # Note from within `simplify_singletons()`
    # TODO: This is not perfect.
    # TODO: Some 3CC artifacts were non-planar but not captured here.

    artifacts["non_planar"] = artifacts["stroke_count"] > artifacts["node_count"]
    a_idx, r_idx = roads.sindex.query(artifacts.geometry.boundary, predicate="overlaps")
    artifacts.iloc[np.unique(a_idx), artifacts.columns.get_loc("non_planar")] = True

    return artifacts


def simplify_singletons(
    artifacts: gpd.GeoDataFrame,
    roads: gpd.GeoDataFrame,
    max_segment_length: float | int = 1,
    compute_coins: bool = True,
    min_dangle_length: float | int = 10,
    eps: float = 1e-4,
    clip_limit: float | int = 2,
    simplification_factor: float | int = 2,
    consolidation_tolerance: float | int = 10,
) -> gpd.GeoDataFrame:
    """Simplification of singleton face artifacts – the first simplification step in
    the procedure detailed in ``simplify.simplify_loop()``.

    This process extracts nodes from network edges before computing and labeling
    face artifacts with a ``{C, E, S}`` typology through ``momepy.COINS`` via the
    constituent road geometries.

    Next, each artifact is iterated over and constituent line geometries are either
    dropped or added in the following order of typologies:
        1. 1 node and 1 continuity group
        2. more than 1 node and 1 or more identical continuity groups
        3. 2 or more nodes and 2 or more continuity groups

    Non-planar geometries are ignored.

    Parameters
    ----------
    artifacts : geopandas.GeoDataFrame
        Face artifact polygons.
    roads : geopandas.GeoDataFrame
        Preprocessed road network data.
    max_segment_length : float | int = 1
        Additional vertices will be added so that all line segments
        are no longer than this value. Must be greater than 0.
        Used in multiple internal geometric operations.
    compute_coins : bool = True
        Flag for computing and labeling artifacts with a ``{C, E, S}`` typology through
        ``momepy.COINS`` via the constituent road geometries.
    min_dangle_length : float | int = 10
        The threshold for determining if linestrings are dangling slivers to be
        removed or not.
    eps : float = 1e-4
        Tolerance epsilon used in multiple internal geometric operations.
    clip_limit : float | int = 2
        Following generation of the Voronoi linework, we clip to fit inside the
        polygon. To ensure we get a space to make proper topological connections
        from the linework to the actual points on the edge of the polygon, we clip
        using a polygon with a negative buffer of ``clip_limit`` or the radius of
        maximum inscribed circle, whichever is smaller.
    simplification_factor : float | int = 2
        The factor by which singles, pairs, and clusters are simplified. The
        ``max_segment_length`` is multiplied by this factor to get the
        simplification epsilon.
    consolidation_tolerance : float | int = 10
        Tolerance passed to node consolidation when generating Voronoi skeletons.

    Returns
    -------
    geopandas.GeoDataFrame
        The road network line data following the singleton procedure.
    """

    # Extract network nodes and relate to artifacts
    nodes, artifacts = _link_nodes_artifacts("singletons", roads, artifacts, eps)

    # Compute number of stroke groups per artifact
    if compute_coins:
        roads, _ = continuity(roads)
    artifacts = _classify_strokes(artifacts, roads)

    # Filter artifacts caused by non-planar intersections
    artifacts = _identify_non_planar(artifacts, roads)

    # Count intersititial nodes (primes)
    _prime_count = artifacts["node_count"] - artifacts[["C", "E", "S"]].sum(axis=1)
    artifacts["interstitial_nodes"] = _prime_count

    # Define the type label
    ces_type = []
    for x in artifacts[["node_count", "C", "E", "S"]].itertuples():
        ces_type.append(f"{x.node_count}{'C' * x.C}{'E' * x.E}{'S' * x.S}")
    artifacts["ces_type"] = ces_type

    # Collect changes
    to_drop: list[int] = []
    to_add: list[int] = []
    split_points: list[shapely.Point] = []

    # Isolate planar artifacts
    planar = artifacts[~artifacts["non_planar"]].copy()
    planar["buffered"] = planar.buffer(eps)
    if artifacts["non_planar"].any():
        logger.debug(f"IGNORING {artifacts.non_planar.sum()} non planar artifacts")

    # Iterate over each singleton planar artifact and simplify based on typology
    for artifact in planar.itertuples():
        n_nodes = artifact.node_count
        n_strokes = artifact.stroke_count
        cestype = artifact.ces_type

        # Get edges relevant for an artifact
        edges = roads.iloc[roads.sindex.query(artifact.buffered, predicate="covers")]

        # Dispatch by typology
        try:
            # 1 node and 1 continuity group
            if (n_nodes == 1) and (n_strokes == 1):
                logger.debug("FUNCTION n1_g1_identical")
                n1_g1_identical(
                    edges,
                    to_drop=to_drop,
                    to_add=to_add,
                    geom=artifact.geometry,
                    max_segment_length=max_segment_length,
                    clip_limit=clip_limit,
                )
            # More than 1 node and 1 or more identical continuity groups
            elif (n_nodes > 1) and (len(set(cestype[1:])) == 1):
                logger.debug("FUNCTION nx_gx_identical")
                nx_gx_identical(
                    edges,
                    geom=artifact.geometry,
                    to_add=to_add,
                    to_drop=to_drop,
                    nodes=nodes,
                    angle=75,
                    max_segment_length=max_segment_length,
                    clip_limit=clip_limit,
                    consolidation_tolerance=consolidation_tolerance,
                )
            # 2 or more nodes and 2 or more continuity groups
            elif (n_nodes > 1) and (len(cestype) > 2):
                logger.debug("FUNCTION nx_gx")
                nx_gx(
                    edges,
                    artifact=artifact,
                    to_drop=to_drop,
                    to_add=to_add,
                    split_points=split_points,
                    nodes=nodes,
                    max_segment_length=max_segment_length,
                    clip_limit=clip_limit,
                    min_dangle_length=min_dangle_length,
                    consolidation_tolerance=consolidation_tolerance,
                )
            else:
                logger.debug("NON PLANAR")
        except Exception as e:
            if DEBUGGING:
                raise e
            warnings.warn(
                f"An error occured at location {artifact.geometry.centroid}. "
                f"The artifact has not been simplified. The original message:\n{e}",
                UserWarning,
                stacklevel=2,
            )

    cleaned_roads = roads.drop(to_drop)
    # split lines on new nodes
    cleaned_roads = split(split_points, roads.drop(to_drop), roads.crs)

    if to_add:
        # Create new roads with fixed geometry.
        # Note: ``to_add`` and ``to_drop`` lists shall be global and
        # this step should happen only once, not for every artifact
        _add_merged = gpd.GeoSeries(to_add).line_merge()
        new = gpd.GeoDataFrame(geometry=_add_merged, crs=roads.crs).explode()
        new = new[~new.normalize().duplicated()].copy()
        new["_status"] = "new"
        new.geometry = new.simplify(max_segment_length * simplification_factor)
        new_roads = pd.concat([cleaned_roads, new], ignore_index=True)
        agg: dict[str, str | typing.Callable] = {"_status": _status}
        for c in cleaned_roads.columns.drop(cleaned_roads.active_geometry_name):
            if c != "_status":
                agg[c] = "first"
        non_empties = new_roads[~(new_roads.is_empty | new_roads.geometry.isna())]
        new_roads = remove_false_nodes(non_empties, aggfunc=agg)

        final = new_roads
    else:
        final = cleaned_roads

    if "coins_group" in final.columns:
        final = final.drop(columns=[c for c in roads.columns if c.startswith("coins_")])
    return final


def simplify_pairs(
    artifacts: gpd.GeoDataFrame,
    roads: gpd.GeoDataFrame,
    max_segment_length: float | int = 1,
    min_dangle_length: float | int = 20,
    clip_limit: float | int = 2,
    simplification_factor: float | int = 2,
    consolidation_tolerance: float | int = 10,
) -> gpd.GeoDataFrame:
    """Simplification of pairs of face artifacts – the second simplification step in
    the procedure detailed in ``simplify.simplify_loop()``.

    This process extracts nodes from network edges before identifying non-planarity
    and cluster information.

    If paired artifacts are present we further classify them as grouped by
    first vs. last instance of duplicated component label, and whether
    or not to be simplified with the clustered process.

    Finally, simplification is performed based on the following order of typologies:
        1. Singletons – merged pairs & first instance (w/o COINS)
        2. Singletons – Second instance – w/ COINS
        3. Clusters

    Parameters
    ----------
    artifacts : geopandas.GeoDataFrame
        Face artifact polygons.
    roads : geopandas.GeoDataFrame
        Preprocessed road network data.
    max_segment_length : float | int = 1
        Additional vertices will be added so that all line segments
        are no longer than this value. Must be greater than 0.
        Used in multiple internal geometric operations.
    min_dangle_length : float | int = 20
        The threshold for determining if linestrings are dangling slivers to be
        removed or not.
    clip_limit : float | int = 2
        Following generation of the Voronoi linework, we clip to fit inside the
        polygon. To ensure we get a space to make proper topological connections
        from the linework to the actual points on the edge of the polygon, we clip
        using a polygon with a negative buffer of ``clip_limit`` or the radius of
        maximum inscribed circle, whichever is smaller.
    simplification_factor : float | int = 2
        The factor by which singles, pairs, and clusters are simplified. The
        ``max_segment_length`` is multiplied by this factor to get the
        simplification epsilon.
    consolidation_tolerance : float | int = 10
        Tolerance passed to node consolidation when generating Voronoi skeletons.

    Returns
    -------
    geopandas.GeoDataFrame
        The road network line data following the pairs procedure.
    """

    # Extract network nodes and relate to artifacts
    nodes, artifacts = _link_nodes_artifacts("pairs", roads, artifacts, None)

    # Compute number of stroke groups per artifact
    roads, _ = continuity(roads)
    artifacts = _classify_strokes(artifacts, roads)

    # Filter artifacts caused by non-planar intersections
    artifacts = _identify_non_planar(artifacts, roads)

    # Identify non-planar clusters
    _id_np = lambda x: sum(artifacts.loc[artifacts["comp"] == x.comp]["non_planar"])  # noqa: E731
    artifacts["non_planar_cluster"] = artifacts.apply(_id_np, axis=1)
    # Subset non-planar clusters and planar artifacts
    np_clusters = artifacts[artifacts.non_planar_cluster > 0]
    artifacts_planar = artifacts[artifacts.non_planar_cluster == 0]

    # Isolate planar artifacts
    _planar_grouped = artifacts_planar.groupby("comp")[artifacts_planar.columns]
    _solutions = _planar_grouped.apply(get_solution, roads=roads)
    artifacts_w_info = artifacts.merge(_solutions, left_on="comp", right_index=True)

    # Isolate non-planar clusters of value 2 – e.g., artifact under highway
    _np_clust_2 = np_clusters["non_planar_cluster"] == 2
    artifacts_under_np = np_clusters[_np_clust_2].dissolve("comp", as_index=False)

    # Determine typology dispatch if artifacts are present
    if not artifacts_w_info.empty:
        agg = {
            "coins_group": "first",
            "coins_end": lambda x: x.any(),
            "_status": _status,
        }
        for c in roads.columns.drop(
            [roads.active_geometry_name, "coins_count"], errors="ignore"
        ):
            if c not in agg:
                agg[c] = "first"

        sol_drop = "solution == 'drop_interline'"
        sol_iter = "solution == 'iterate'"

        # Determine artifacts and road edges to drop
        _to_drop = artifacts_w_info.drop_duplicates("comp").query(sol_drop).drop_id
        _drop_roads = roads.drop(_to_drop.dropna().values)

        # Re-run node cleaning on subset of fresh road edges
        roads_cleaned = remove_false_nodes(
            _drop_roads,
            aggfunc=agg,
        )

        # Isolate drops to create merged pairs
        merged_pairs = artifacts_w_info.query(sol_drop).dissolve("comp", as_index=False)

        # Sort artifacts by their node count low-to-high
        sorted_node_count = artifacts_w_info.sort_values("node_count", ascending=False)

        # Isolate artifacts to process as singletons – first instance
        _1st = sorted_node_count.query(sol_iter).drop_duplicates("comp", keep="first")
        _planar_clusters = np_clusters[~np_clusters["non_planar"]]
        _1st = pd.concat([_1st, _planar_clusters], ignore_index=True)

        # Isolate artifacts to process as singletons – last instance
        _2nd = sorted_node_count.query(sol_iter).drop_duplicates("comp", keep="last")

        # Isolate artifacts to process as clusters
        for_skeleton = artifacts_w_info.query("solution == 'skeleton'")

    # Otherwise instantiate artifact containers as empty
    else:
        merged_pairs = pd.DataFrame()
        _1st = pd.DataFrame()
        _2nd = pd.DataFrame()
        for_skeleton = pd.DataFrame()
        roads_cleaned = roads

    # Generate counts of COINs groups for edges
    coins_count = (
        roads_cleaned.groupby("coins_group", as_index=False)
        .geometry.count()
        .rename(columns={"geometry": "coins_count"})
    )
    roads_cleaned = roads_cleaned.merge(coins_count, on="coins_group", how="left")

    # Add under non-planars to cluster dispatcher
    if not artifacts_under_np.empty:
        for_skeleton = pd.concat([for_skeleton, artifacts_under_np])

    # Dispatch singleton simplifier
    if not merged_pairs.empty or not _1st.empty:
        # Merged pairs & first instance – w/o COINS
        roads_cleaned = simplify_singletons(
            pd.concat([merged_pairs, _1st]),
            roads_cleaned,
            max_segment_length=max_segment_length,
            clip_limit=clip_limit,
            compute_coins=False,
            min_dangle_length=min_dangle_length,
            simplification_factor=simplification_factor,
            consolidation_tolerance=consolidation_tolerance,
        )
        # Second instance – w/ COINS
        if not _2nd.empty:
            roads_cleaned = simplify_singletons(
                _2nd,
                roads_cleaned,
                max_segment_length=max_segment_length,
                clip_limit=clip_limit,
                compute_coins=True,
                min_dangle_length=min_dangle_length,
                simplification_factor=simplification_factor,
                consolidation_tolerance=consolidation_tolerance,
            )

    # Dispatch cluster simplifier
    if not for_skeleton.empty:
        roads_cleaned = simplify_clusters(
            for_skeleton,
            roads_cleaned,
            max_segment_length=max_segment_length,
            simplification_factor=simplification_factor,
            min_dangle_length=min_dangle_length,
            consolidation_tolerance=consolidation_tolerance,
        )

    return roads_cleaned


def simplify_clusters(
    artifacts: gpd.GeoDataFrame,
    roads: gpd.GeoDataFrame,
    max_segment_length: float | int = 1,
    eps: float = 1e-4,
    simplification_factor: float | int = 2,
    min_dangle_length: float | int = 20,
    consolidation_tolerance: float | int = 10,
) -> gpd.GeoDataFrame:
    """Simplification of clusters of face artifacts – the third simplification step in
    the procedure detailed in ``simplify.simplify_loop()``.

    This process extracts nodes from network edges before iterating over each
    cluster artifact and performing simplification.

    Parameters
    ----------
    artifacts : geopandas.GeoDataFrame
        Face artifact polygons.
    roads : geopandas.GeoDataFrame
        Preprocessed road network data.
    max_segment_length : float | int = 1
        Additional vertices will be added so that all line segments
        are no longer than this value. Must be greater than 0.
        Used in multiple internal geometric operations.
    eps : float = 1e-4
        Tolerance epsilon used in multiple internal geometric operations.
    simplification_factor : float | int = 2
        The factor by which singles, pairs, and clusters are simplified. The
        ``max_segment_length`` is multiplied by this factor to get the
        simplification epsilon.
    min_dangle_length : float | int = 20
        The threshold for determining if linestrings are dangling slivers to be
        removed or not.
    consolidation_tolerance : float | int = 10
        Tolerance passed to node consolidation when generating Voronoi skeletons.

    Returns
    -------
    geopandas.GeoDataFrame
        The road network line data following the clusters procedure.
    """

    # Get nodes from the network
    nodes = momepy.nx_to_gdf(momepy.node_degree(momepy.gdf_to_nx(roads)), lines=False)

    # Collect changes
    to_drop: list[int] = []
    to_add: list[int] = []

    for _, artifact in artifacts.groupby("comp"):
        # Get artifact cluster polygon
        cluster_geom = artifact.union_all()
        # Get edges relevant for an artifact
        edges = roads.iloc[
            roads.sindex.query(cluster_geom, predicate="intersects")
        ].copy()

        # Clusters of 2 or more nodes and 2 or more continuity groups
        nx_gx_cluster(
            edges=edges,
            cluster_geom=cluster_geom,
            nodes=nodes,
            to_drop=to_drop,
            to_add=to_add,
            eps=eps,
            max_segment_length=max_segment_length,
            min_dangle_length=min_dangle_length,
            consolidation_tolerance=consolidation_tolerance,
        )

    cleaned_roads = roads.drop(to_drop)

    # Create new roads with fixed geometry.
    # Note: ``to_add`` and ``to_drop`` lists shall be global and
    # this step should happen only once, not for every artifact
    new = gpd.GeoDataFrame(geometry=to_add, crs=roads.crs)
    new["_status"] = "new"
    new["geometry"] = new.line_merge().simplify(
        max_segment_length * simplification_factor
    )
    new_roads = pd.concat([cleaned_roads, new], ignore_index=True).explode()
    agg: dict[str, str | typing.Callable] = {"_status": _status}
    for c in new_roads.columns.drop(new_roads.active_geometry_name):
        if c != "_status":
            agg[c] = "first"
    new_roads = remove_false_nodes(
        new_roads[~new_roads.is_empty], aggfunc=agg
    ).drop_duplicates("geometry")

    return new_roads


def get_type(edges: gpd.GeoDataFrame, shared_edge: int) -> str:
    """Classify artifact edges according to the ``{C, E, S}``
    schema when considering solutions for pairs of artifacts.

    Parameters
    ----------
    edges : geopandas.GeoDataFrame
        Artifact edges in consideration.
    shared_edge : int
        The index location of the shared edge of the pair.

    Returns
    -------
    str
        Classification for an edge in ``{C, E, S}``.
    """

    if (  # Roundabout special case
        edges["coins_group"].nunique() == 1
        and edges.shape[0] == edges["coins_count"].iloc[0]
    ):
        return "S"

    all_ends = edges[edges["coins_end"]]
    mains = edges[~edges["coins_group"].isin(all_ends["coins_group"])]
    shared = edges.loc[shared_edge]

    if shared_edge in mains.index:
        return "C"

    if shared["coins_count"] == (edges["coins_group"] == shared["coins_group"]).sum():
        return "S"

    return "E"


def get_solution(group: gpd.GeoDataFrame, roads: gpd.GeoDataFrame) -> pd.Series:
    """Determine the solution for paired planar artifacts.

    Parameters
    ----------
    group : geopandas.GeoDataFrame
        Dissolved group of connected planar artifacts.
    roads : geopandas.GeoDataFrame
        Road network data.

    Returns
    -------
    pandas.Series
        The determined solution and edge to drop.
    """

    def _relate(loc: int) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """Isolate intersecting & covering road geometries."""
        _geom = group.geometry.iloc[loc]
        _roads = roads.iloc[roads.sindex.query(_geom, predicate="intersects")]
        _covers = _roads.iloc[_roads.sindex.query(_geom, predicate="covers")]
        return _roads, _covers

    cluster_geom = group.union_all()

    roads_a, covers_a = _relate(0)
    roads_b, covers_b = _relate(1)

    # Find the road segment that is contained within the cluster geometry
    shared = roads.index[roads.sindex.query(cluster_geom, predicate="contains")]

    if shared.empty or covers_a.empty or covers_b.empty:
        return pd.Series({"solution": "non_planar", "drop_id": None})

    shared = shared.item()

    if (np.invert(roads_b.index.isin(covers_a.index)).sum() == 1) or (
        np.invert(roads_a.index.isin(covers_b.index)).sum() == 1
    ):
        return pd.Series({"solution": "drop_interline", "drop_id": shared})

    seen_by_a = get_type(covers_a, shared)
    seen_by_b = get_type(covers_b, shared)

    if seen_by_a == "C" and seen_by_b == "C":
        return pd.Series({"solution": "iterate", "drop_id": shared})

    if seen_by_a == seen_by_b:
        return pd.Series({"solution": "drop_interline", "drop_id": shared})

    return pd.Series({"solution": "skeleton", "drop_id": shared})


def simplify_network(
    roads: gpd.GeoDataFrame,
    *,
    max_segment_length: float | int = 1,
    min_dangle_length: float | int = 20,
    clip_limit: float | int = 2,
    simplification_factor: float | int = 2,
    consolidation_tolerance: float | int = 10,
    artifact_threshold: None | float | int = None,
    artifact_threshold_fallback: None | float | int = None,
    area_threshold_blocks: float | int = 1e5,
    isoareal_threshold_blocks: float | int = 0.5,
    area_threshold_circles: float | int = 5e4,
    isoareal_threshold_circles_enclosed: float | int = 0.75,
    isoperimetric_threshold_circles_touching: float | int = 0.9,
    eps: float = 1e-4,
    exclusion_mask: None | gpd.GeoSeries = None,
    predicate: str = "intersects",
) -> gpd.GeoDataFrame:
    """Top-level workflow for simplifying networks. The input raw road network data,
    which must be in a projected coordinate reference system and is expected to be in
    meters, is first preprocessed (topological corrections & node consolidation) before
    two iterations of artifact detection and simplification.

    Each iteration of the simplification procedure which includes (1.) the removal
    of false nodes; (2.) face artifact classification; and (3.) the line-based
    simplification of face artifacts in the order of single artifacts, pairs of
    artifacts, clusters of artifacts.

    For further information on face artifact detection and extraction
    see :cite:`fleischmann2023`.

    Parameters
    ----------
    roads : geopandas.GeoDataFrame
        Raw road network data. This input *must* be in a projected coordinate reference
        system and *should* be in meters. All defaults arguments assume meters.
    max_segment_length : float | int = 1
        Additional vertices will be added so that all line segments
        are no longer than this value. Must be greater than 0.
        Used in multiple internal geometric operations.
    min_dangle_length : float | int
        The threshold for determining if linestrings are dangling slivers to be
        removed or not.
    clip_limit : float | int = 2
        Following generation of the Voronoi linework, we clip to fit inside the
        polygon. To ensure we get a space to make proper topological connections
        from the linework to the actual points on the edge of the polygon, we clip
        using a polygon with a negative buffer of ``clip_limit`` or the radius of
        maximum inscribed circle, whichever is smaller.
    simplification_factor : float | int = 2
        The factor by which singles, pairs, and clusters are simplified. The
        ``max_segment_length`` is multiplied by this factor to get the
        simplification epsilon.
    consolidation_tolerance : float | int = 10
        Tolerance passed to node consolidation when generating Voronoi skeletons.
    artifact_threshold : None | float | int = None
        When ``artifact_threshold`` is passed, the computed value from
        ``momepy.FaceArtifacts.threshold`` is not used in favor of the
        given value. This is useful for small networks where artifact
        detection may fail or become unreliable.
    artifact_threshold_fallback : None | float | int = None
        If artifact threshold detection fails, this value is used as a fallback.
    area_threshold_blocks : float | int = 1e5
        This is the first threshold for detecting block-like artifacts whose
        Face Artifact Index (see :cite:`fleischmann2023`) is above the value
        passed in ``artifact_threshold``.
        If a polygon has an area below ``area_threshold_blocks``, *and*
        is of elongated shape (see also ``isoareal_threshold_blocks``),
        *and* touches at least one polygon that has already been classified as artifact,
        then it will be classified as an artifact.
    isoareal_threshold_blocks : float | int = 0.5
        This is the second threshold for detecting block-like artifacts whose
        Face Artifact Index (see :cite:`fleischmann2023`) is above the value
        passed in ``artifact_threshold``. If a polygon has an isoareal quotient
        below ``isoareal_threshold_blocks`` (see ``esda.shape.isoareal_quotient``),
        i.e., if it has an elongated shape; *and* it has a sufficiently small area
        (see also ``area_threshold_blocks``), *and* if it touches at least one
         polygon that has already been detected as an artifact,
        then it will be classified as an artifact.
    area_threshold_circles : float | int = 5e4
        This is the first threshold for detecting circle-like artifacts whose
        Face Artifact Index (see :cite:`fleischmann2023`) is above the value
        passed in ``artifact_threshold``. If a polygon has an area below
        ``area_threshold_circles``, *and* one of the following 2 cases is given:
        (a) the polygon is touched, but not enclosed by polygons already classified
        as artifacts, *and* with an isoperimetric quotient
        (see ``esda.shape.isoperimetric_quotient``)
        above ``isoperimetric_threshold_circles_touching``, i.e., if its shape
        is close to circular; or (b) the polygon is fully enclosed by polygons
        already classified as artifacts, *and* with an isoareal quotient
        above
        ``isoareal_threshold_circles_enclosed``, i.e., if its shape is
        close to circular; then it will be classified as an artifact.
    isoareal_threshold_circles_enclosed : float | int = 0.75
        This is the second threshold for detecting circle-like artifacts whose
        Face Artifact Index (see :cite:`fleischmann2023`) is above the value
        passed in ``artifact_threshold``. If a polygon has a sufficiently small
        area (see also ``area_threshold_circles``), *and* the polygon is
        fully enclosed by polygons already classified as artifacts,
        *and* its isoareal quotient (see ``esda.shape.isoareal_quotient``)
        is above the value passed to ``isoareal_threshold_circles_enclosed``,
        i.e., if its shape is close to circular;
        then it will be classified as an artifact.
    isoperimetric_threshold_circles_touching : float | int = 0.9
        This is the third threshold for detecting circle-like artifacts whose
        Face Artifact Index (see :cite:`fleischmann2023`)
        is above the value passed in ``artifact_threshold``.
        If a polygon has a sufficiently small area
        (see also ``area_threshold_circles``), *and* the polygon is touched
        by at least one polygon already classified as artifact,
        *and* its isoperimetric quotient (see ``esda.shape.isoperimetric_quotient``)
        is above the value passed to ``isoperimetric_threshold_circles_touching``,
        i.e., if its shape is close to circular;
        then it will be classified as an artifact.
    eps : float = 1e-4
        Tolerance epsilon used in multiple internal geometric operations.
    exclusion_mask : None | geopandas.GeoSeries = None
        Polygons used to determine face artifacts to exclude from returned output.
    predicate : str = 'intersects'
        The spatial predicate used to exclude face artifacts from returned output.

    Returns
    -------
    geopandas.GeoDataFrame
        The final, simplified road network line data.

    Notes
    -----
    As is noted above, the input network data must be in a projected coordinate
    reference system and is expected to be in meters. However, it may be possible to
    work with network data projected in feet if all default arguments are adjusted.
    """

    # NOTE: this keeps attributes but resets index
    roads = fix_topology(roads, eps=eps)
    # Merge nearby nodes (up to double of distance used in skeleton).
    # NOTE: this drops attributes and resets index
    roads = consolidate_nodes(roads, tolerance=max_segment_length * 2.1)

    # Identify artifacts
    artifacts, threshold = get_artifacts(
        roads,
        threshold=artifact_threshold,
        threshold_fallback=artifact_threshold_fallback,
        area_threshold_blocks=area_threshold_blocks,
        isoareal_threshold_blocks=isoareal_threshold_blocks,
        area_threshold_circles=area_threshold_circles,
        isoareal_threshold_circles_enclosed=isoareal_threshold_circles_enclosed,
        isoperimetric_threshold_circles_touching=isoperimetric_threshold_circles_touching,
        exclusion_mask=exclusion_mask,
        predicate=predicate,
    )

    # Loop 1
    new_roads = simplify_loop(
        roads,
        artifacts,
        max_segment_length=max_segment_length,
        min_dangle_length=min_dangle_length,
        clip_limit=clip_limit,
        simplification_factor=simplification_factor,
        consolidation_tolerance=consolidation_tolerance,
        eps=eps,
    )

    # This is potentially fixing some minor erroneous edges coming from Voronoi
    new_roads = induce_nodes(new_roads, eps=eps)
    new_roads = new_roads[~new_roads.geometry.normalize().duplicated()].copy()

    # Identify artifacts based on the first loop network
    artifacts, _ = get_artifacts(
        new_roads,
        threshold=threshold,
        threshold_fallback=artifact_threshold_fallback,
        area_threshold_blocks=area_threshold_blocks,
        isoareal_threshold_blocks=isoareal_threshold_blocks,
        area_threshold_circles=area_threshold_circles,
        isoareal_threshold_circles_enclosed=isoareal_threshold_circles_enclosed,
        isoperimetric_threshold_circles_touching=isoperimetric_threshold_circles_touching,
        exclusion_mask=exclusion_mask,
        predicate=predicate,
    )

    # Loop 2
    final_roads = simplify_loop(
        new_roads,
        artifacts,
        max_segment_length=max_segment_length,
        min_dangle_length=min_dangle_length,
        clip_limit=clip_limit,
        simplification_factor=simplification_factor,
        consolidation_tolerance=consolidation_tolerance,
        eps=eps,
    )

    # This is potentially fixing some minor erroneous edges coming from Voronoi
    final_roads = induce_nodes(final_roads, eps=eps)
    final_roads = final_roads[~final_roads.geometry.normalize().duplicated()].copy()

    return final_roads


def simplify_loop(
    roads: gpd.GeoDataFrame,
    artifacts: gpd.GeoDataFrame,
    max_segment_length: float | int = 1,
    min_dangle_length: float | int = 20,
    clip_limit: float | int = 2,
    simplification_factor: float | int = 2,
    consolidation_tolerance: float | int = 10,
    eps: float = 1e-4,
) -> gpd.GeoDataFrame:
    """Perform an iteration of the simplification procedure which includes:
        1. Removal of false nodes
        2. Artifact classification
        3. Simplifying artifacts:
            - Single artifacts
            - Pairs of artifacts
            - Clusters of artifacts

    Parameters
    ----------
    roads : geopandas.GeoDataFrame
        Raw road network data.
    artifacts : geopandas.GeoDataFrame
        Face artifact polygons.
    max_segment_length : float | int = 1
        Additional vertices will be added so that all line segments
        are no longer than this value. Must be greater than 0.
        Used in multiple internal geometric operations.
    min_dangle_length : float | int = 20
        The threshold for determining if linestrings are dangling slivers to be
        removed or not.
    clip_limit : float | int = 2
        Following generation of the Voronoi linework, we clip to fit inside the
        polygon. To ensure we get a space to make proper topological connections
        from the linework to the actual points on the edge of the polygon, we clip
        using a polygon with a negative buffer of ``clip_limit`` or the radius of
        maximum inscribed circle, whichever is smaller.
    simplification_factor : float | int = 2
        The factor by which singles, pairs, and clusters are simplified. The
        ``max_segment_length`` is multiplied by this factor to get the
        simplification epsilon.
    consolidation_tolerance : float | int = 10
        Tolerance passed to node consolidation when generating Voronoi skeletons.
    eps : float = 1e-4
        Tolerance epsilon used in multiple internal geometric operations.

    Returns
    -------
    geopandas.GeoDataFrame
        The road network line data following 1 iteration of simplification.
    """

    # Remove edges fully within the artifact (dangles).
    _, r_idx = roads.sindex.query(artifacts.geometry, predicate="contains")
    # Dropping may lead to new false nodes – drop those
    roads = remove_false_nodes(roads.drop(roads.index[r_idx]))

    # Filter singleton artifacts
    rook = graph.Graph.build_contiguity(artifacts, rook=True)

    # Keep only those artifacts which occur as isolates,
    # e.g. artifacts that are not part of a larger intersection
    singles = artifacts.loc[artifacts.index.intersection(rook.isolates)].copy()

    # Filter doubles
    artifacts["comp"] = rook.component_labels
    counts = artifacts["comp"].value_counts()
    doubles = artifacts.loc[artifacts["comp"].isin(counts[counts == 2].index)].copy()

    # Filter clusters
    clusters = artifacts.loc[artifacts["comp"].isin(counts[counts > 2].index)].copy()

    if not singles.empty:
        # NOTE: this drops attributes
        roads = simplify_singletons(
            singles,
            roads,
            max_segment_length=max_segment_length,
            simplification_factor=simplification_factor,
            consolidation_tolerance=consolidation_tolerance,
        )
    if not doubles.empty:
        roads = simplify_pairs(
            doubles,
            roads,
            max_segment_length=max_segment_length,
            min_dangle_length=min_dangle_length,
            clip_limit=clip_limit,
            simplification_factor=simplification_factor,
            consolidation_tolerance=consolidation_tolerance,
        )
    if not clusters.empty:
        roads = simplify_clusters(
            clusters,
            roads,
            max_segment_length=max_segment_length,
            simplification_factor=simplification_factor,
            eps=eps,
            min_dangle_length=min_dangle_length,
            consolidation_tolerance=consolidation_tolerance,
        )

    if "coins_group" in roads.columns:
        roads = roads.drop(columns=[c for c in roads.columns if c.startswith("coins_")])
    return roads
